# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@terkin.org>
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3
import json
from copy import copy
from urllib.parse import urlsplit, urlencode
from terkin import logging
from terkin.model import DataFrame
from terkin.util import to_base64, format_exception, get_device_id, urlparse, dformat, get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()

#log.setLevel(logging.DEBUG)


class TelemetryManager:
    """Manage a number of telemetry adapters."""

    def __init__(self):
        self.adapters = []
        self.errors_seen = {}
        self.failure_count = {}

    def add_adapter(self, adapter):
        """

        :param adapter:

        """
        self.adapters.append(adapter)

    def transmit(self, dataframe: DataFrame):
        """

        :param data:

        """
        outcomes = {}
        for adapter in self.adapters:

            # Dispatch transmission to telemetry adapter.
            outcome = adapter.transmit(dataframe)

            # Todo: Propagate last error message into outcome and obtain here.
            channel = adapter.channel_uri
            outcomes[channel] = outcome

        return outcomes


class TelemetryAdapter:
    """
    Telemetry node client: Network participant API
    Todo: Implement exponential backoff instead of MAX_FAILURES.
    """

    MAX_FAILURES = 3

    def __init__(self, device=None, target=None):

        self.device = device
        self.target = target

        self.interface = self.target.get('interface')
        self.base_uri = self.target['endpoint']
        self.address = self.target.get('address', {})
        self.address['base_uri'] = self.target['endpoint']
        self.data_more = self.target.get('data', {})

        # TODO: Move default value deeper into the framework here?
        self.format = self.target.get('format', TelemetryClient.FORMAT_JSON)
        self.content_encoding = self.target.get('content_encoding')

        self.channel_uri = None
        self.client = None

        self.topology_name = self.target.get('topology')
        self.topology = None

        self.offline = False
        self.failure_count = 0

    def setup(self):
        """ """

        # Resolve designated topology.
        self.topology = self.topology_factory()
        log.info('Telemetry channel topology: %s', self.topology.name)

        # Compute telemetry channel URI.
        self.channel_uri = self.format_uri()
        log.info('Telemetry channel URI: %s', self.channel_uri)

        # Resolve designated telemetry client.
        self.client = self.client_factory()

    def client_factory(self):
        """ """
        client = TelemetryClient(interface=self.interface,
                                 uri=self.channel_uri,
                                 format=self.format,
                                 content_encoding=self.content_encoding,
                                 uri_suffixes=self.topology.uri_suffixes,
                                 settings=self.target.get('settings'),
                                 networking=self.device.networking)
        return client

    def topology_factory(self):
        """ """
        return TelemetryTopologyFactory(name=self.topology_name, adapter=self).create()

    def format_uri(self, **kwargs):
        """

        :param kwargs:

        """
        data = copy(self.address)
        data.update(kwargs)
        return self.topology.uri_template.format(**data)

    def transmit(self, dataframe: DataFrame):
        """

        :param dataframe:

        """

        self.device.watchdog.feed()

        if not self.is_online():
            # Todo: Suppress this message after a while or reduce interval.
            message = 'Adapter is offline, skipping telemetry to {}'.format(self.channel_uri)
            log.warning(message)
            return False

        # Transform into egress telemetry payload
        # using the designated encoder.
        try:
            self.transform(dataframe)
        except Exception as ex:
            log.exc(ex, 'Transmission transform for topology "%s" failed', self.topology_name)
            return False

        try:
            outcome = self.client.transmit(dataframe)
            self.reset_errors()
            return outcome

        except Exception as ex:
            self.record_error()
            message = 'Telemetry to {} failed'.format(self.channel_uri)
            if self.offline:
                log.warning(message)
            else:
                log.exc(ex, message)

        return False

    def transform(self, dataframe: DataFrame):
        """
        :param dataframe:
        """

        # Add predefined "data_more" to telemetry message.
        dataframe.data_in.update(self.data_more)

        # Run data through designated encoder, if given.
        if hasattr(self.topology, 'encode'):
            encoder = self.topology.encode
            if encoder is not None and callable(encoder):
                encoder(dataframe)

    def is_online(self):
        """ """

        # Short-cut online/offline state.
        return True

        # Todo: Re-enable online/offline state.
        #return self.failure_count < self.MAX_FAILURES

    def record_error(self):
        """ """
        self.failure_count += 1

    def reset_errors(self):
        """ """
        self.failure_count = 0


class CSVTelemetryAdapter(TelemetryAdapter):
    """Telemetry node client: Network participant API.

    This will make your node talk CSV.


    """

    def __init__(self, *args, **kwargs):
        TelemetryAdapter.__init__(self, *args, **kwargs)
        self.format = TelemetryClient.FORMAT_CSV

    def transmit(self, data, **kwargs):
        """

        :param data:
        :param kwargs:

        """
        uri = self.format_uri(**kwargs)
        log.info('Telemetry channel URI for CSV: %s', uri)
        return self.client.transmit(data, uri=uri, serialize=False)


class TelemetryClient:
    """
    A flexible telemetry data client wrapping access to
    different transport adapters and serialization mechanisms.
    """

    TRANSPORT_HTTP = 'http'
    TRANSPORT_MQTT = 'mqtt'
    TRANSPORT_LORA = 'lora'

    FORMAT_URLENCODED = 'urlencoded'
    FORMAT_JSON = 'json'
    FORMAT_CSV = 'csv'

    FORMAT_CAYENNELPP_HIVEEYES = 'lpp-hiveeyes'
    FORMAT_CAYENNELPP_RATRACK = 'lpp-ratrack'

    CONTENT_ENCODING_IDENTITY = 'identity'
    CONTENT_ENCODING_BASE64 = 'base64'

    def __init__(self, interface, uri, format, content_encoding=None, uri_suffixes=None, settings=None, networking=None):

        log.info('Starting Terkin TelemetryClient')
        self.interface = interface
        self.uri = uri

        self.transport = None
        self.handlers = {}

        self.format = format
        self.content_encoding = content_encoding
        self.uri_suffixes = uri_suffixes or {}
        self.settings = settings or {}

        self.networking = networking

        self.scheme, self.netloc, self.path, self.query, self.fragment = urlsplit(self.uri)

        if self.scheme in ['http', 'https']:
            self.transport = TelemetryClient.TRANSPORT_HTTP

        elif self.scheme in ['mqtt']:
            self.transport = TelemetryClient.TRANSPORT_MQTT

        elif self.scheme in ['lora']:
            self.transport = TelemetryClient.TRANSPORT_LORA

    def serialize(self, dataframe: DataFrame):
        """

        :param data: defined in terkin/model.py

        moves & reformats data from dataframe.data_out to dataframe.payload_out

        """

        # Serialize payload.
        if self.format == TelemetryClient.FORMAT_URLENCODED:
            payload = urlencode(dataframe.data_out)

        elif self.format == TelemetryClient.FORMAT_JSON:
            payload = json.dumps(dataframe.data_out)

        elif self.format == TelemetryClient.FORMAT_CAYENNELPP_HIVEEYES:
            from terkin.telemetry.formatter import to_cayenne_lpp_hiveeyes
            payload = to_cayenne_lpp_hiveeyes(dataframe)

        elif self.format == TelemetryClient.FORMAT_CAYENNELPP_RATRACK:
            from terkin.telemetry.formatter import to_cayenne_lpp_ratrack
            payload = to_cayenne_lpp_ratrack(dataframe)

        elif self.format == TelemetryClient.FORMAT_CSV:
            raise NotImplementedError('Serialization format "CSV" not implemented yet')

        else:
            raise ValueError('Unknown serialization format "{}"'.format(self.format))

        # Apply content encoding.
        if self.content_encoding in (None, self.CONTENT_ENCODING_IDENTITY):
            pass

        elif self.content_encoding == self.CONTENT_ENCODING_BASE64:
            payload = to_base64(payload)

        dataframe.payload_out = payload

    def transmit(self, dataframe: DataFrame, uri=None, serialize=True):
        """
        :param dataframe:
        :param uri:  (Default value = None)
        :param serialize:  (Default value = True)
        """

        # Submit telemetry data using HTTP POST request
        # Serialization: x-www-form-urlencoded

        # Compute target URI.
        if uri:
            real_uri = uri
        else:
            real_uri = self.uri

        suffix = self.uri_suffixes.get(self.transport, '').format(**self.__dict__)
        real_uri = real_uri + suffix

        # Resolve handler by URI.
        handler = self.get_handler(real_uri)

        # Prepare dataframe for egress.
        dataframe.payload_out = dataframe.data_out
        if serialize:
            self.serialize(dataframe)

        # Submit dataframe to handler.
        return handler.send(dataframe)

    def get_handler(self, uri):
        """
        :param uri:
        """

        if uri in self.handlers:
            return self.handlers[uri]

        if self.transport == TelemetryClient.TRANSPORT_HTTP and self.interface == 'gprs':
            handler = TelemetryTransportHTTPOverGPRS(self.networking.gprs_manager, uri, self.format)

        elif self.transport == TelemetryClient.TRANSPORT_HTTP:
            handler = TelemetryTransportHTTP(uri, self.format)

        elif self.transport == TelemetryClient.TRANSPORT_MQTT:
            handler = TelemetryTransportMQTT(uri, self.format)

        elif self.transport == TelemetryClient.TRANSPORT_LORA:
            handler = TelemetryTransportLORA(self.networking.lora_manager, self.settings)

        else:
            raise ValueError('Unknown telemetry transport "{}"'.format(self.transport))

        self.handlers[uri] = handler

        return handler


class TelemetryTransportHTTP:
    """ """

    def __init__(self, uri, format):

        self.uri = uri
        self.format = format

        self.scheme, self.netloc, self.path, self.query, self.fragment = urlsplit(self.uri)

        self.content_type = None
        self.resolve_content_type()

    def resolve_content_type(self):
        """ """

        if self.format == TelemetryClient.FORMAT_URLENCODED:
            self.content_type = 'application/x-www-form-urlencoded'

        elif self.format == TelemetryClient.FORMAT_JSON:
            self.content_type = 'application/json'

        elif self.format == TelemetryClient.FORMAT_CSV:
            self.content_type = 'text/csv'

        else:
            raise ValueError('Unknown serialization format for TelemetryTransportHTTP: {}'.format(format))

    def send(self, dataframe: DataFrame):
        """
        Submit telemetry data using HTTP POST request

        :param dataframe:
        """
        log.info('Sending HTTP request to %s', self.uri)
        log.info('HTTP payload: %s', dataframe.payload_out)

        import urequests
        response = urequests.post(self.uri, data=dataframe.payload_out.encode(), headers={'Content-Type': self.content_type})
        if response.status_code in [200, 201]:
            return True
        else:
            message = 'HTTP request failed: {} {}\n{}'.format(response.status_code, response.reason, response.content)
            raise TelemetryTransportError(message)


class TelemetryTransportHTTPOverGPRS(TelemetryTransportHTTP):

    def __init__(self, gprs_manager, uri, format):
        super().__init__(uri, format)
        self.gprs_manager = gprs_manager

        log.info('Telemetry transport: JSON over HTTP over GPRS')

    def send(self, dataframe: DataFrame):
        """
        Submit telemetry data using HTTP POST request over GPRS.

        :param dataframe:
        """
        log.info('Sending HTTP request over GPRS APN %s to %s', self.gprs_manager.gprs_settings.get('apn'), self.uri)
        log.info('Payload:     %s', dataframe.payload_out)

        response = self.gprs_manager.send_request(self.uri, dataframe.payload_out, self.content_type)
        if response.status_code in [200, 201]:
            return True
        else:
            # TODO: Implement ``response.reason``.
            #message = 'HTTP request failed: {} {}\n{}'.format(response.status_code, response.reason, response.content)
            message = 'HTTP request failed: {}\n{}'.format(response.status_code, response.content)
            raise TelemetryTransportError(message)


class TelemetryTransportLORA:
    """ """

    lora_socket = None

    def __init__(self, lora_adapter, settings):

        log.info('Telemetry transport: CayenneLPP over LoRaWAN/TTN')

        self.lora_adapter = lora_adapter
        self.settings = settings or {}
        self.size = self.settings.get('size', 12)
        self.datarate = self.settings.get('datarate', 0)

    def send(self, dataframe: DataFrame):
        """
        :param dataframe:
        """

        import binascii

        if platform_info.vendor == platform_info.MICROPYTHON.Pycom:
            import pycom
            nvram_get = pycom.nvs_get
            nvram_set = pycom.nvs_set
            nvram_erase = pycom.nvs_erase
        elif platform_info.vendor == platform_info.MICROPYTHON.Vanilla:
            import esp32
            nvram_get = esp32.nvs_get
            nvram_set = esp32.nvs_set
            nvram_erase = esp32.nvs_erase

        if self.lora_adapter is None:
            log.error('LoRa not enabled or no antenna attached')
            return False

        self.lora_adapter.ensure_connectivity()

        # Default payload.
        payload = dataframe.payload_out

        # Clean up payload from sensor data if pause command was received on last uplink.
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
            try:
                _pause = nvram_get('pause')
            except:
                _pause = None

            if _pause is None or _pause == 0:
                payload = dataframe.payload_out + binascii.unhexlify(b'000100')
            elif _pause == 1:
                payload = binascii.unhexlify(b'000101')

        log.info('[LoRa] Uplink payload (hex): %s', binascii.hexlify(payload).decode())

        # Send payload
        success = False
        try:
            log.info('[LoRa] Sending payload...')
            outcome = self.lora_adapter.send(payload)
            log.info('[LoRa] Sent %s bytes', outcome)
            success = True

        except:
            log.exception('[LoRa] Transmission failed')
            return False

        # Receive downlink message and save integer value for
        # 1) deep sleep interval in minutes and
        # 2) pausing payload submission (1=true,0=false)
        # survives power cycle, reset and deep sleep
        rx, port = self.lora_adapter.receive()

        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
            if port == 1:
                sleep = int.from_bytes(rx, "big")
                if sleep == 0:
                    # use value from settings file
                    log.info('[LoRa] Received "reset deep sleep interval" command, erasing from NVRAM.')
                    try:
                        nvram_erase('deepsleep')
                    except:
                        pass
                else:
                    # Use deepsleep interval received via LoRa.
                    log.info('[LoRa] Received "set deep sleep interval" command, will sleep for %s minutes.', sleep)
                    nvram_set('deepsleep', sleep)

            elif port == 2:
                pause = int.from_bytes(rx, "big")
                log.info('[LoRa] Received "pause payload submission" command: %s', bool(pause))
                nvram_set('pause', pause)

            else:
                log.info('[LoRa] No downlink message processed')

            success = success and True

        else:
            log.warning('[LoRa] Processing downlink messages not implemented for Dragino')

        return success


class TelemetryTransportMQTT:
    """
    MQTT transport for Terkin Telemetry.

    This is currently based on the "Pycom MicroPython MQTT module" just called ``mqtt.py``.
    https://github.com/pycom/pycom-libraries/blob/master/lib/mqtt/mqtt.py

    Originally, this was based on the "umqtt.robust" library::

        micropython -m upip install micropython-umqtt.robust micropython-umqtt.simple


    """

    connections = {}

    def __init__(self, uri, format):

        log.info('Telemetry transport: MQTT over TCP')

        # Addressing.
        self.uri = uri
        self.format = format

        # v1
        #self.scheme, self.netloc, self.path, self.query, self.fragment = urlsplit(self.uri)

        # v2
        self.target = urlparse(self.uri)

        # Todo: Refactor `get_device_id` somehow better.
        self.client_id = 'terkin.{}'.format(get_device_id())

        # Status flags.
        self.defunct = False
        self.defunctness_reported = False

        # TODO: Start connecting to MQTT broker here already?
        #       Maybe we should defer this to the point where the
        #       first transmission will require it.
        self.ensure_connection()

    def ensure_connection(self):
        """ """
        # Create one MQTTAdapter instance per target host:port.
        if self.target.netloc not in self.connections:
            # TODO: Add more parameters to MQTTAdapter here.
            log.info('Connecting to MQTT broker at {} with username {}. '
                     'client_id={}'.format(self.target.hostname, self.target.username, self.client_id))
            try:
                self.connections[self.target.netloc] = MQTTAdapter(self.client_id,
                                                                   self.target.hostname,
                                                                   username=self.target.username,
                                                                   password=self.target.password)
            except Exception:
                log.warning('Connecting to MQTT broker at {} '
                            'with username {} failed'.format(self.target.hostname, self.target.username))
                # Todo: Re-enable defunctness
                #self.defunct = True

        return self.connections.get(self.target.netloc)

    def get_connection(self):
        """ """
        return self.ensure_connection()

    def send(self, dataframe: DataFrame):
        """
        :param dataframe: defined in terkin/model.py

        sends data in dataframe.payload_out per MQTT to a server
        """

        # Evaluate and handle defunctness.
        if self.defunct:
            if not self.defunctness_reported:
                log.error('MQTT transport is defunct, please scan log '
                          'output for previous error messages.')
                self.defunctness_reported = True
            return False

        # Derive MQTT topic string from URI path component.
        topic = self.target.path.lstrip('/')
        if not topic:
            message = 'Empty MQTT topic, please configure MQTT URI with path component or topology with address'
            raise TelemetryTransportError(message)

        # Use payload from request.
        payload = dataframe.payload_out

        # Reporting.
        log.info('MQTT topic:   %s', topic)
        prettify_log = False
        if prettify_log:
            log.info('MQTT payload:\n\ns', dformat(payload, indent=48))
        else:
            log.info('MQTT payload: %s', payload)

        try:
            connection = self.get_connection()
            connection.publish(topic, payload)

        except TelemetryAdapterError as ex:
            message = 'Protocol adapter not connected: {}'.format(format_exception(ex))
            raise TelemetryTransportError(message)

        return True


class MQTTAdapter:
    """
    MQTT adapter wrapping the lowlevel MQTT driver.
    Handles a single connection to an MQTT broker.

    TODO: Try to make this module reasonably compatible again
          by becoming an adapter for different implementations.
          E.g., what about Paho?


    """

    def __init__(self, client_id, server, port=0, username=None, password=None):

        # TODO: Add more parameters: keepalive=0, ssl=False, ssl_params={}
        self.client_id = client_id
        self.server = server
        self.port = port
        self.username = username
        self.password = password

        # Transport driver.
        self.driver_class = None
        self.load_driver()

        # Connection instance.
        self.connection = None

        # Status flags.
        self.connected = False

    def load_driver(self):
        """Load MQTT driver module"""

        # TODO: Create abstract MQTT client factory to account for different implementations.
        try:
            from umqtt import MQTTClient
            self.driver_class = MQTTClient

        except Exception as ex:
            log.exc(ex, 'Loading MQTT module failed')
            raise

    def ensure_connection(self):
        """Conditionally connect to MQTT broker, if not connected already"""
        if not self.connected:
            self.connect()

    def connect(self):
        """Connect to MQTT broker"""
        try:
            log.info('Connecting to MQTT broker at {} with username {}'.format(self.server, self.username))
            self.connection = self.driver_class(self.client_id, self.server, port=self.port, user=self.username, password=self.password)
            self.connection.DEBUG = True
            self.connection.connect()
            self.connected = True
            log.info('Connecting to MQTT broker at %s succeeded', self.connection.addr)

        except Exception as ex:
            # FIXME: Evaluate exception. "MQTTException: 5" means "Authentication/Authorization failed".
            self.connected = False
            message = 'Connecting to MQTT broker at {} failed: {}'.format(self.server, format_exception(ex))
            log.exc(ex, message)
            raise TelemetryAdapterError(message)

        return self.connected

    def publish(self, topic, payload, retain=False, qos=1):
        """

        :param topic:
        :param payload:
        :param retain:  (Default value = False)
        :param qos:  (Default value = 1)

        """

        # Try to (re-)connect to MQTT broker.
        self.ensure_connection()

        if not self.connected:
            message = 'No MQTT connectivity, skipping telemetry'
            log.warning(message)
            raise TelemetryAdapterError(message)

        try:
            # TODO: Make qos level configurable.
            self.connection.publish(topic, payload, retain=retain, qos=qos)

        except OSError as ex:

            message = 'MQTT publishing failed'
            log.exc(ex, message)

            # Signal connection error in order to reconnect on next submission attempt.
            # [Errno 104] ECONNRESET
            # [Errno 113] ECONNABORTED
            try:
                if ex.errno in [104, 113]:
                    self.connected = False
            except AttributeError:
                print('Exception has no "errno" attribute', dir(ex))
                self.connected = False

            message = '{}: {}'.format(message, ex)
            raise TelemetryAdapterError(message)


class TelemetryTopology:
    """Define **how** to communicate using Telemetry."""

    NONE = 'null'
    MQTTKIT = 'mqttkit'
    BEEP_BOB = 'beep-bob'


class TelemetryTopologyFactory:
    """ """

    def __init__(self, name=None, adapter=None):
        self.name = name
        self.adapter = adapter

    def create(self):
        """ """

        if self.name is None or self.name == TelemetryTopology.NONE:
            return IdentityTopology()

        elif self.name == TelemetryTopology.BEEP_BOB:
            return BeepBobTopology(adapter=self.adapter)

        elif self.name == TelemetryTopology.MQTTKIT:
            return MqttKitTopology()

        else:
            raise KeyError('Configured topology "{}" unknown'.format(self.name))


class IdentityTopology:
    """Apply no kind of transformation to telemetry payload
    and provide Base for derivatives of me.


    """
    uri_template = u'{base_uri}'
    uri_suffixes = None

    @property
    def name(self):
        """ """
        return self.__class__.__name__

    def encode(self, dataframe: DataFrame):
        """

        :param dataframe: defined in terkin/model.py

        move data_in to data_out (why?)

        """
        dataframe.data_out = dataframe.data_in


class BeepBobTopology(IdentityTopology):
    """Define how to communicate with BEEP for BOB.

    https://en.wikipedia.org/wiki/Bebop

    - https://beep.nl/
    - https://github.com/beepnl/BEEP
    - https://hiverize.org/
    - https://bee-observer.org/


    """
    uri_template = u'{base_uri}'
    uri_suffixes = None

    def __init__(self, adapter=None):
        self.adapter = adapter
        self.settings = self.adapter.device.settings

    def encode(self, dataframe: DataFrame):
        """Encode telemetry data matching the BEEP-BOB interface.

        https://gist.github.com/vkuhlen/51f7968266659f37d076bd66d57cdbbd
        https://github.com/Hiverize/FiPy/blob/master/logger/beep.py

        Example::

            {
                't': 22.66734,
                'h': 52.41612,
                'p': 1002.334,
                'weight': 10.0,
                't_i_1': 23.1875,
                't_i_2': 23.125,
                't_i_3': 23.125,
                't_i_4': 23.1875
            }

        :param data:

        """

        # Rename all fields to designated BEEP-BOB fields.
        egress_data = {}
        mapping = self.settings.get('telemetry.maps.bee-observer', self.settings.get('sensor_telemetry_map'))
        mapping['key'] = 'key'
        for sensor_field, telemetry_field in mapping.items():
            sensor_field = sensor_field.lower()
            if sensor_field in dataframe.data_in:
                egress_data[telemetry_field] = dataframe.data_in[sensor_field]

        dataframe.data_out = egress_data


class MqttKitTopology(IdentityTopology):
    """This defines how to communicate in WAN scenarios having a decent
    number of devices rolled out. While this would cover even earth-scale
    addressing scenarios, it will also give you peace of mind in smaller
    setups like multi-project or multi-tenant scenarios. Even for single
    users, the infinite number of available channels is very convenient
    for ad hoc operation scenarios.

    - https://getkotori.org/
    - https://getkotori.org/docs/applications/mqttkit.html

    - https://hiveeyes.org/
    - https://hiveeyes.org/docs/system/acquisition/


    """
    uri_template = u'{base_uri}/{realm}/{network}/{gateway}/{node}'
    uri_suffixes = {
        TelemetryClient.TRANSPORT_HTTP: '/data',
        TelemetryClient.TRANSPORT_MQTT: '/data.{format}',
    }


class TelemetryTransportError(Exception):
    """ """
    pass


class TelemetryAdapterError(Exception):
    """ """
    pass



from terkin import logging
from terkin.driver import hx711_sensor
from terkin.sensor import SensorManager, AbstractSensor

log = logging.getLogger(__name__)

def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create HX711 sensor object decorator for temperature compensation.

    Compensated load is calculated with the following formula:
        compensated_load = current_load + (temperature_offset - current_temp) * temp_compensation_factor

    :param sensor_manager:
    :param sensor_info:

    :return: decorator_object
    """

    log.info('Creating  hx711Sensor_object...')
    hx711Sensor_object = hx711_sensor.includeme(sensor_manager, sensor_info)
    log.info('Creating  HX711TempCompensationSensor...')
    decorator_object = HX711TempCompensationSensor(
        sensor_manager=sensor_manager, hx711Sensor_object=hx711Sensor_object, settings=sensor_info)
    
    decorator_object.family = hx711Sensor_object.family
    decorator_object.register_parameter('temperature_sensor_id', sensor_info['temperature_sensor_id'])
    decorator_object.register_parameter('temperature_offset', float(sensor_info['temperature_offset']))
    decorator_object.register_parameter('temp_compensation_factor', float(sensor_info['temp_compensation_factor']))

    log.info('HX711TempCompensationSensor created.')
    return decorator_object


class HX711TempCompensationSensor(AbstractSensor):
    """ Decorates HX7711Sensor class to add temperature compensation features
    """

    hx711Sensor_object = None
    sensor_manager = None

    def __init__(self, sensor_manager: SensorManager, hx711Sensor_object, settings=None):
        super().__init__(settings=settings)
        self.hx711Sensor_object = hx711Sensor_object
        self.sensor_manager = sensor_manager

    def start(self):
        """ """
        self.hx711Sensor_object.start()

    def read(self):
        """ """
        temperature = self.read_temperature()
        load_data = self.read_hx711_data()

        decorated_data = {}
        for key, value in load_data.items():
            if (key.startswith('weight')):
                decorated_data[key] = self.calculate_compensated_load(value, temperature)
            else:
                decorated_data[key] = value

        return decorated_data

    def power_on(self):
        """ """
        self.hx711Sensor_object.power_on()

    def power_off(self):
        """ """
        self.hx711Sensor_object.power_off()

    def calculate_compensated_load(self, weight, temperature):
        """ """
        tempOffset = self.parameter['temperature_offset']
        compensationFactor = self.parameter['temp_compensation_factor']

        compensated_weight = weight + (temperature - tempOffset) * compensationFactor
        log.info("weight compensation changed weight from '{}' to '{}'".format(weight, compensated_weight))
        return compensated_weight

    def read_temperature(self):
        """ """
        temp_sensor_id = self.parameter['temperature_sensor_id']
        log.info('reading temperature from {}'.format(temp_sensor_id))
        temp_sensor = self.sensor_manager.get_sensor_by_id(temp_sensor_id)
        
        temp_sensor_reading = temp_sensor.read()
        temp_sensor_data = temp_sensor_reading.data

        # Evaluate sensor outcome.
        if temp_sensor_data is None or temp_sensor_data is AbstractSensor.SENSOR_NOT_INITIALIZED:
            raise Exception("no temperature available")

        temperature = None
        for key, value in temp_sensor_data.items():
            if (key.startswith('temperature')):
                temperature = value
                break

        log.info('read temperature is {}'.format(temperature))
        if (temperature == None):
            raise Exception("no temperature available")

        return temperature

    def read_hx711_data(self):
        """ """
        log.info('reading sensor data from decorated hx711 sensor...')
        load_sensor_data = self.hx711Sensor_object.read()

        return load_sensor_data



from terkin import logging
from terkin.model import DataFrame

log = logging.getLogger(__name__)


def to_cayenne_lpp_hiveeyes(dataframe: DataFrame):
    """
    Serialize dataframe to binary CayenneLPP format.

    :param dataframe:
    """

    from cayennelpp import LppFrame
    frame = LppFrame()

    channel = {}
    channel['temp']   = 10
    channel['volt']   = 2
    channel['pres']   = 1
    channel['scal']   = 6

    # TODO: Iterate ``dataframe.readings`` to get more metadata from sensor configuration.
    # It is a list of ``SensorReading`` instances, each having a ``sensor`` and ``data`` attribute.

    # log.info('dataframe.data_out.items : %s', dataframe.data_out.items())

    for key, value in dataframe.data_out.items():

        #log.debug('-' * 42)

        # TODO: Maybe implement different naming conventions.
        name = key.split("_")[0]
        #     channel = 0

        if "system" in name:
            if "voltage" in name:
                # treat voltage as system sensors
                # put battery voltage to channel 0
                # put solar voltage to channel 1
                # put other voltages to channel 2++
                if "battery" in name:
                    chan = 0
                elif "solar" in name:
                    chan = 1
                else:
                    chan = channel['volt']
                    channel['volt'] += 1
                frame.add_voltage(chan, value)
            elif "temperature" in name:
                frame.add_temperature(0, value)
        elif "i2c" in name:
            # assume BME280 on I2C bus as outside sensor and assign the variables to channel 5
            if "temperature" in name:
                frame.add_temperature(5, value)
            elif "humidity" in name:
                frame.add_humidity(5, value)
            elif "pressure" in name:
                frame.add_barometer(5, value)
        elif "weight" in name:
            # treat weight as outside sensor and assign to channel 5-9
            # Channel 5 for single reading (weight.0))
            # Channel 6-9 for multiple readings (weight.[1-4])
            if name == "weight.0":
                chan = 5
            else:
                chan = channel['scal']
                channel['scal'] += 1
            frame.add_load(chan, value)
        elif "onewire" in name:
            # assume DS18B20 as inside sensors and assign to channel 10++
            if "temperature" in name:
                frame.add_temperature(channel['temp'], value)
                channel['temp'] += 1
        elif "analog-output" in name:
            frame.add_analog_output(channel, value)
        elif "analog-input" in name:
            frame.add_analog_input(channel, value)
        elif "digital-input" in name:
            frame.add_digital_input(channel, value)
        elif "digital_output" in name:
            frame.add_digital_output(channel, value)
        elif "illuminance" in name:
            frame.add_luminosity(channel, value)
        elif "barometer" in name:
            frame.add_barometer(channel, value)
        elif "presence" in name:
            frame.add_presence(channel, value)
        elif "accelerometer" in name:
            frame.add_accelerometer(channel, value)
        elif "gyrometer" in name:
            frame.add_gyrometer(channel, value)
        #elif "gps" in name:
        #    frame.add_gps(channel, value)


        # TODO: Fork cayenneLPP and implement load cell telemetry.
        # TODO: Add load encoder as ID 122 (3322)
        # http://openmobilealliance.org/wp/OMNA/LwM2M/LwM2MRegistry.html#extlabel
        # http://www.openmobilealliance.org/tech/profiles/lwm2m/3322.xml

        # elif False and "load" in name:
        #     frame.add_load(channel, value)

        # TODO: Map memfree and other baseline sensors appropriately.

        else:
            # TODO: raise Exception here?
            log.info('[CayenneLPP] Sensor type "{}" not found in CayenneLPP'.format(name))

    return frame.bytes()


def to_cayenne_lpp_ratrack(dataframe: DataFrame):
    """
    Serialize dataframe to binary CayenneLPP format.

    :param dataframe:
    """

    from cayennelpp import LppFrame
    frame = LppFrame()

    channel = {}
    channel['temp']   = 10

    if ('alt' or 'altitude') and ('lat' or 'longitude') and ('lon' or 'longitude') in dataframe.data_out.keys():
        log.info('GOT GPS!')
        frame.add_gps(0, dataframe.data_out.get('lat'), dataframe.data_out.get('lon'), dataframe.data_out.get('alt'))

    if 'batterie_volt' in dataframe.data_out.keys():
        frame.add_analog_input(14, dataframe.data_out.get('batterie_volt'))

    for key, value in dataframe.data_out.items():

        name = key

        if "system" in name:
            pass
        elif "i2c" in name:
            # assume BME280 on I2C bus as outside sensor and assign the variables to channel 5
            if "temperature" in name:
                frame.add_temperature(5, value)
            elif "humidity" in name:
                frame.add_humidity(5, value)
            elif "pressure" in name:
                frame.add_barometer(5, value)
        elif "onewire" in name:
            # assume DS18B20 as inside sensors and assign to channel 10++
            if "temperature" in name:
                frame.add_temperature(channel['temp'], value)
                channel['temp'] += 1
        else:
            # TODO: raise Exception here?
            #log.info('[CayenneLPP] Sensor type "{}" not found in CayenneLPP'.format(name))
            hu1 = 2

    return frame.bytes()



# -*- coding: utf-8 -*-
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3
import os
import sys
import uio
import copy
import collections

from umal import PlatformInfo, GenericChronometer


def to_base64(bytes):
    """Encode bytes to base64 encoded string

    :param bytes: 

    """
    # TODO: Move to ``util.py``.
    import base64
    return base64.encodebytes(bytes).decode().rstrip()


def format_exception(ex):
    """

    :param ex: 

    """
    return '{}: {}'.format(ex.__class__.__name__, str(ex))


def get_device_id():
    """ 
    MAC address of device if supported.
    """
    import machine
    from ubinascii import hexlify
    return hexlify(machine.unique_id()).decode()


class URI:
    """ """
    def __str__(self):
        return str(self.__dict__)


def urlparse(url, scheme='', allow_fragments=True):
    """Parse a URL into 6 components:
    <scheme>://<netloc>/<path>;<params>?<query>#<fragment>

    :param url: 
    :param scheme:  (Default value = '')
    :param allow_fragments:  (Default value = True)
    :returns: Note that we don't break the components up in smaller bits
    (e.g. netloc is a single string) and we don't expand % escapes.

    """
    from urllib.parse import uses_params, urlsplit, _splitparams, _coerce_args, ParseResult
    url, scheme, _coerce_result = _coerce_args(url, scheme)
    splitresult = urlsplit(url, scheme, allow_fragments)
    scheme, netloc, url, query, fragment = splitresult
    if scheme in uses_params and ';' in url:
        url, params = _splitparams(url)
    else:
        params = ''
    result = ParseResult(scheme, netloc, url, params, query, fragment)

    # FIXME: Appropriately parse netloc into parts using the real ``urlparse``.
    # However, Pycom MicroPython currently lacks the ``rpartition`` method on string objects.

    uri = URI()
    for fieldname in ['scheme', 'netloc', 'path', 'params', 'query', 'fragment']:
        key = fieldname
        value = getattr(result, fieldname)
        #print(key, value)
        setattr(uri, key, value)

    setattr(uri, 'hostname', uri.netloc)
    setattr(uri, 'username', None)
    setattr(uri, 'password', None)

    #print('URI-1:', uri)

    # Manually parse credentials from netloc.
    # Fixme: Improve urlparse to do the same.
    if '@' in uri.netloc:
        credentials, hostname = uri.netloc.split('@')
        username, password = credentials.split(':')

        setattr(uri, 'hostname', hostname)
        setattr(uri, 'username', username)
        setattr(uri, 'password', password)

    #print('URI-2:', uri)

    return uri


def str_grouper(n, iterable):
    """

    :param n: 
    :param iterable: 

    """
    # https://stackoverflow.com/questions/11006702/elegant-format-for-a-mac-address-in-python-3-2/11006779#11006779
    args = [iter(iterable)] * n
    for part in zip(*args):  # itertools.izip in 2.x for efficiency.
        yield "".join(part)


def format_mac_address(mac_address):
    """

    :param mac_address: 

    """
    return ":".join(str_grouper(2, mac_address)).lower()


def dformat(data, indent=0):
    """

    :param data: 
    :param indent:  (Default value = 0)

    """
    padding = ' ' * indent
    from uio import StringIO
    buffer = StringIO()
    for key in sorted(data.keys()):
        value = data[key]
        buffer.write('{}{}: {}\n'.format(padding, key, value))
    return buffer.getvalue()


def ddformat(data, indent=0):
    """

    :param data: 
    :param indent:  (Default value = 0)

    """
    padding = ' ' * indent
    from uio import StringIO
    buffer = StringIO()
    for key in sorted(data.keys()):
        item = data[key]
        value = item['value']
        text = item.get('description', '')
        buffer.write('{}{:<35}{:>25} {:>25}\n'.format(padding, key, value, text))
    return buffer.getvalue()


def _flatten(input_obj, key_prefix, separator='_'):
    """
    Flatten any type of python object into one-level dict object.
    https://github.com/evamayerova/python-flatten/blob/master/flatten_to_dict/flatten_to_dict.py

    :param input_obj: param key_prefix:
    :param key_prefix: 
    :param separator:  (Default value = '_')

    """
    new_dict = {}
    if type(input_obj) is dict:
        for key, value in input_obj.items():
            if type(value) is dict or type(value) is list or type(value) is tuple:
                new_key = key_prefix + key + separator
                new_dict.update(_flatten(input_obj[key], new_key, separator))
            else:
                new_dict[key_prefix + key] = value
    elif type(input_obj) is list or type(input_obj) is tuple:
        for nr, item in enumerate(input_obj):
            new_key = key_prefix + str(nr) + separator
            new_dict.update(_flatten(item, new_key, separator))
    else:
        new_dict[key_prefix[:-1]] = input_obj

    return new_dict


def flatten(input_object, separator='_'):
    """Flatten any object into one-level dict representation.

    :param input_object: 
    :param separator:  (Default value = '_')
    :returns: Flattened object in dict representation.
    Examples:
    https://github.com/evamayerova/python-flatten/blob/master/flatten_to_dict/flatten_to_dict.py
    :rtype: dict

    >>> a = {"f": ["a", "b"], "b": {"x": [1, 2, 3]}, "i": 1}
        >>> flatten(a)
        {
            'f_0': 'a',
            'f_1': 'b',
            'b_x_0': 1,
            'b_x_1': 2,
            'b_x_2': 3,
            'i': 1
         }
    """
    return _flatten(input_object, key_prefix="", separator=separator)


class gc_disabled:
    """Context manager to temporarily disable the garbage collector.
    
    Please be aware this piece does not account for thread safety in any way.
    
    https://community.hiveeyes.org/t/timing-things-on-micropython-for-esp32/2329
    https://bugs.python.org/issue31356
    
    Synopsis::
    
        with gc_disabled():
            # Do something that needs realtime guarantees
            # such as a pair trade, robotic braking, etc.
            run_some_timing_critical_stuff()


    """

    def __enter__(self):
        import gc
        gc.disable()
        return self

    def __exit__(self, *exc_details):
        # exc_info: (<class 'NameError'>, NameError("name 'asdf' is not defined",), None)
        received_exc = exc_details[0] is not None
        import gc
        gc.enable()
        if received_exc:
            raise exc_details[1]


def file_remove(fn: str) -> None:
    """Try to remove a file if it exists.
    
    From logging.handlers.

    :param fn: str: 

    """
    #print('os.remove:', fn)
    try:
        os.remove(fn)
    except OSError:
        pass


def file_exists(fn: str) -> bool:
    """

    :param fn: str: 

    """
    #print('os.stat:', fn)
    try:
        os.stat(fn)
        return True
    except OSError:
        pass
    return False


def ensure_directory(path: str) -> None:
    """

    :param path: str: 

    """
    #print('file_exists')
    if file_exists(path):
        return True
    return os.mkdir(path)


def get_last_stacktrace():
    """ """
    buf = uio.StringIO()
    exc = sys.exc_info()[1]
    sys.print_exception(exc, buf)
    return buf.getvalue()


def random_from_crypto():
    """ """
    platform_info = get_platform_info()
    if platform_info.vendor == platform_info.MICROPYTHON.Pycom:
        # https://forum.pycom.io/topic/1378/solved-how-to-get-random-number-in-a-range/6
        # https://github.com/micropython/micropython-lib/blob/master/random/random.py
        import crypto
        r = crypto.getrandbits(32)
    else:
        import urandom
        r = urandom.getrandbits(32)
    return ((r[0] << 24) + (r[1] << 16) + (r[2] << 8) + r[3]) / 4294967295.0


def randint(a, b):
    """

    :param a: 
    :param b: 

    """
    return random_from_crypto() * (b - a) + a


def backoff_time(n, minimum=1, maximum=600):
    """

    :param n: 
    :param minimum:  (Default value = 1)
    :param maximum:  (Default value = 600)

    """
    # https://en.wikipedia.org/wiki/Exponential_backoff
    # https://cloud.google.com/storage/docs/exponential-backoff
    # https://cloud.google.com/iot/docs/how-tos/exponential-backoff
    # https://github.com/GoogleCloudPlatform/python-docs-samples/blob/master/iot/api-client/mqtt_example/cloudiot_mqtt_example.py
    # https://stackoverflow.com/questions/27438273/exponential-backoff-time-sleep-with-random-randint0-1000-1000
    random_delta = randint(0, 1000) / 1000.0
    delay = min(max((2 ** n) + random_delta, minimum), maximum)
    return delay


class Stopwatch:
    """ """

    def __init__(self):
        self.chronometer = GenericChronometer()
        self.begin = self.chronometer.read()

    def elapsed(self):
        """ """
        return self.chronometer.read() - self.begin

    def reset(self):
        """ """
        self.chronometer.reset()


class PycomChronometer:
    """A chronometer implemented with Pycom MicroPython.
    https://docs.pycom.io/firmwareapi/pycom/machine/timer/


    """

    def __init__(self):
        from machine import Timer
        self.chrono = Timer.Chrono()
        self.chrono.start()

    def read(self):
        """ """
        return self.chrono.read()

    def reset(self):
        """ """
        self.chrono.reset()


class Eggtimer:
    """ """

    def __init__(self, duration):
        self.duration = duration
        self.stopwatch = Stopwatch()

    def expired(self):
        """ """
        return self.stopwatch.elapsed() >= self.duration


def get_platform_info() -> PlatformInfo:
    """ """
    from __main__ import bootloader
    return bootloader.platform_info


def dict_merge(dct, merge_dct, add_keys=True):
    """Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    
    This version will return a copy of the dictionary and leave the original
    arguments untouched.
    
    The optional argument ``add_keys``, determines whether keys which are
    present in ``merge_dict`` but not ``dct`` should be included in the
    new dict.

    :param dct: dict
    :param merge_dct: dict
    :param add_keys: bool (Default value = True)
    :returns: dict: updated dict
    
    Resources:
        https://gist.github.com/angstwad/bf22d1822c38a92ec0a9

    """
    dct = dct.copy()
    if not add_keys:
        merge_dct = {
            k: merge_dct[k]
            for k in set(dct).intersection(set(merge_dct))
        }

    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict)
                and isinstance(merge_dct[k], collections.Mapping)):
            dct[k] = dict_merge(dct[k], merge_dct[k], add_keys=add_keys)
        else:
            dct[k] = merge_dct[k]

    return dct


# Copyright Ferry Boender, released under the MIT license.
# https://www.electricmonk.nl/log/2017/05/07/merging-two-python-dictionaries-by-deep-updating/
def deepupdate(target, src):
    """Deep update target dict with src
    For each k,v in src: if k doesn't exist in target, it is deep copied from
    src to target. Otherwise, if v is a list, target[k] is extended with
    src[k]. If v is a set, target[k] is updated with v, If v is a dict,
    recursively deep-update it.
    
    Examples:

    :param target: 
    :param src: 

    >>> t = {'name': 'Ferry', 'hobbies': ['programming', 'sci-fi']}
    >>> deepupdate(t, {'hobbies': ['gaming']})
    >>> print t
    {'name': 'Ferry', 'hobbies': ['programming', 'sci-fi', 'gaming']}
    """
    for k, v in src.items():
        if type(v) == list:
            if not k in target:
                target[k] = copy.deepcopy(v)
            else:
                target[k].extend(v)
        elif type(v) == dict:
            if not k in target:
                target[k] = copy.deepcopy(v)
            else:
                deepupdate(target[k], v)
        elif type(v) == set:
            if not k in target:
                target[k] = v.copy()
            else:
                target[k].update(v.copy())
        else:
            target[k] = copy.copy(v)



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
"""
Universal MicroPython Application Loader.
"""


class McuFamily:
    """
    Constants for designating different MCU families.
    """

    # STMicroelectronics
    # STM32 32-bit Arm Cortex MCUs.
    # https://en.wikipedia.org/wiki/STM32
    # https://www.st.com/en/microcontrollers-microprocessors/stm32-32-bit-arm-cortex-mcus.html
    STM32 = 1

    # Espressif
    # System on a chip microcontrollers with integrated Wi-Fi and dual-mode Bluetooth.
    # https://en.wikipedia.org/wiki/ESP32
    # https://www.espressif.com/en/products/hardware/esp32/overview
    ESP32 = 2


class MicroPythonVendor:
    """
    Constants for designating different vendors of the MicroPython platform.
    """

    # Vanilla/Genuine MicroPython.
    # https://github.com/micropython/micropython
    Vanilla = 1

    # Pycom MicroPython.
    # https://github.com/pycom/pycom-micropython-sigfox
    Pycom = 2

    # RaspberryPi.
    RaspberryPi = 3

    # Odroid.
    Odroid = 4


class PlatformInfo:
    """
    Platform info object bundling the combination of MCU family
    and MicroPython vendor to assist in platform switching
    mechanics throughout the program.
    """

    MCU = McuFamily
    MICROPYTHON = MicroPythonVendor

    def __init__(self):

        self.mcu = None
        self.vendor = None
        self.micropython_version = None
        self.device_name = None

        self.resolve_platform()

    def resolve_platform(self):
        """ 
        Resolve MCU and vendor of this platform.
        """

        import sys

        self.micropython_version = sys.implementation.version

        if sys.platform == 'esp32':
            self.mcu = McuFamily.ESP32
            self.vendor = MicroPythonVendor.Vanilla

        if sys.platform in ['WiPy', 'LoPy', 'LoPy4', 'SiPy', 'GPy', 'FiPy']:
            self.mcu = McuFamily.ESP32
            self.vendor = MicroPythonVendor.Pycom

        if sys.platform in ['pyboard']:
            self.mcu = McuFamily.STM32
            self.vendor = MicroPythonVendor.Vanilla

        try:
            linux_firmware = open('/sys/firmware/devicetree/base/model').read()
            if 'Raspberry' in linux_firmware:
                self.mcu = McuFamily.STM32
                self.vendor = MicroPythonVendor.RaspberryPi
            elif 'Odroid' in linux_firmware:
                self.mcu = McuFamily.STM32
                self.vendor = MicroPythonVendor.Odroid
        except:
            pass

        self.device_name = sys.platform


class ApplicationInfo:
    """ 
    Collects various information. Most importantly transports the settings for further use in the Datalogger.
    """

    def __init__(self, name=None, version=None, settings=None, application=None, platform_info: PlatformInfo = None):
        self.name = name
        self.version = version

        self.platform_info = platform_info

        self.settings = settings
        self.application = application

    @property
    def fullname(self):
        """ """
        return '{} {}'.format(self.name, self.version)


class MicroPythonBootloader:
    """ 
    Very basic tasks for running the system.
    """

    def __init__(self):
        self.platform_info = PlatformInfo()
        self.duty_chrono = GenericChronometer()

    def extend_syspath(self):
        """
        Extend Python module search path.

        Dependency modules are shipped through the "dist-packages" folder.
        Please populate this folder appropriately as shown above before
        expecting anything to work.

        The default ``PYTHONPATH`` as found on the different platforms is:
        - Vanilla MicroPython: ['', '/lib']
        - Pycom MicroPython: ['', '/flash', '/flash/lib']
        """
        import sys

        # Extend by path containing frozen modules.
        paths = ['lib', 'dist-packages', 'lib-mpy']
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            paths = ['/flash/{}'.format(path) for path in paths]

        elif self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
            paths = ['/{}'.format(path) for path in paths]

        elif self.platform_info.vendor == self.platform_info.MICROPYTHON.RaspberryPi:
            paths = ['./dist-packages', './src/lib']

        else:
            print('[umal]    ERROR: MicroPython platform not supported:', self.platform_info.vendor)
            sys.exit(1)

        # Amend module search path.
        sys.path.clear()
        sys.path.extend(paths)

        # According to @poesel, empty directory designates frozen modules.
        # We want to put these at the end in order to be able to override
        # any module we would like to hack upon.
        sys.path.extend([''])

        print('[umal]    INFO: Python module search path is:', sys.path)


class GenericChronometer:
    """
    A millisecond chronometer implemented with vanilla MicroPython.
    https://micropython.readthedocs.io/en/latest/pyboard/tutorial/timer.html#making-a-microsecond-counter
    """

    def __init__(self):
        import time
        self.start = time.ticks_ms()

    def read(self):
        """ """
        import time
        return time.ticks_diff(time.ticks_ms(), self.start) / 1000.0

    def reset(self):
        """ """
        import time
        self.start = time.ticks_ms()



import os
import time
import binascii

from terkin import logging

log = logging.getLogger(__name__)


class LinuxSysfsOneWireBus:
    """
    dtoverlay w1-gpio gpiopin=4 pullup=0

    cat /sys/bus/w1/devices/w1_bus_master1/w1_master_slaves
    28-0517c482cfff
    """
    def __init__(self, sysfs):
        self.sysfs = sysfs
        self.roms = []
        log.info('Starting LinuxSysfsOneWireBus on {}'.format(self.sysfs))

    def scan(self):
        with open(os.path.join(self.sysfs, 'w1_master_slaves'), 'r') as f:
            items = f.readlines()
            self.roms = []
            for item in items:
                # Decode 1-Wire device address on sysfs like "28-0517c482cfff".
                prefix, address = item.split('-')
                rom = bytes.fromhex(prefix) + bytes.fromhex(address)
                self.roms.append(rom)
            return self.roms


class LinuxSysfsDS18B20:
    """
    cat /sys/bus/w1/devices/w1_bus_master1/28-0517c482cfff/w1_slave
    36 01 4b 46 7f ff 0c 10 9a : crc=9a YES
    36 01 4b 46 7f ff 0c 10 9a t=19375
    """
    def __init__(self, bus):
        self.bus = bus

    def convert_temp(self):
        # Noop
        pass

    def read_temp(self, address):
        address = insert(binascii.hexlify(address).decode(), '-', 2)
        log.info('DS18B20 read_temp: %s', address)
        device_file = os.path.join(self.bus.sysfs, address, 'w1_slave')
        log.info('DS18B20 device_file: %s', device_file)
        temp_c, temp_f = self.read_temp_raw(device_file)
        return temp_c

    def read_temp_raw(self, device_file):
        # https://www.waveshare.com/wiki/Raspberry_Pi_Tutorial_Series:_1-Wire_DS18B20_Sensor

        f = open(device_file, 'r')
        lines = f.readlines()
        f.close()

        # Analyze if the last 3 characters are 'YES'.
        while lines[0].strip()[-3:] != 'YES':
            time.sleep(0.2)
            lines = self.read_temp_raw(device_file)
        # Find the index of 't=' in a string.
        equals_pos = lines[1].find('t=')
        if equals_pos != -1:
            # Read the temperature .
            temp_string = lines[1][equals_pos+2:]
            temp_c = float(temp_string) / 1000.0
            temp_f = temp_c * 9.0 / 5.0 + 32.0
            return temp_c, temp_f


def insert(source_str, insert_str, pos):
    # https://stackoverflow.com/a/25978101
    return source_str[:pos] + insert_str + source_str[pos:]



# -*- coding: utf-8 -*-
# (c) 2020 Andreas Motl <andreas.motl@terkin.org>
# (c) 2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2020 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3
import os
import sys
import logging

log = logging.getLogger()


def configure_pythonpath(search_paths):
    sys.path = aslist(search_paths) + sys.path


def patch_system():
    # Fix: RuntimeError: Click will abort further execution because Python 3
    #      was configured to use ASCII as encoding for the environment.
    os.environ['LANG'] = 'en_US.UTF-8'

    # Make environment compatible with CPython.
    log.info('Setting up CPython compatibility layer')
    from terkin_cpython.compat import monkeypatch_cpython
    monkeypatch_cpython()


def setup_logging():
    """
    Configure CPython logger.
    """
    import sys
    import logging
    log_format = '%(asctime)-15s [%(name)-30s] %(levelname)-7s: %(message)s'
    logging.basicConfig(
        format=log_format,
        stream=sys.stderr,
        level=logging.INFO)


def start_bootloader():
    """
    Invoke bootloader.
    """
    import sys
    from umal import MicroPythonBootloader
    bootloader = MicroPythonBootloader()
    sys.modules['__main__'].bootloader = bootloader
    return bootloader


def load_settings(settings_file):

    log.info('Loading settings from "{}"'.format(settings_file))

    # Sanity checks.
    if not os.path.exists(settings_file):
        raise FileNotFoundError('Settings file "{}" not found'.format(settings_file))

    # https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path/67692#67692
    import importlib.util
    spec = importlib.util.spec_from_file_location("settings", settings_file)
    settings = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(settings)
    return settings


def aslist(obj, sep=None, strip=True):
    """
    # (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
    # Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
    """
    if isinstance(obj, str):
        lst = obj.split(sep)
        if strip:
            lst = [v.strip() for v in lst]
        return lst
    elif isinstance(obj, (list, tuple)):
        return obj
    elif obj is None:
        return []
    else:
        return [obj]



# -*- coding: utf-8 -*-
# (c) 2017-2019 Andreas Motl <andreas@terkin.org>
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging

log = logging.getLogger(__name__)


class Watchdog:
    """The watchdog timer (WDT) is used to restart the system when the
    application crashes and ends up into a non recoverable state. After
    enabling, the application must "feed" the watchdog periodically to
    prevent the timeout from expiring and resetting the system.
    
    https://docs.pycom.io/firmwareapi/pycom/machine/wdt/

    """

    def __init__(self, device=None, settings=None):
        self.device = device
        self.settings = settings
        self.wdt = None
        self.suspended = False

        self.enabled = self.settings.get('main.watchdog.enabled', False)
        self.timeout = self.settings.get('main.watchdog.timeout', 10000)

    def start(self):
        """ 
        Start the watchdog.
        """
        if not self.enabled:
            log.info('Skipping watchdog timer (WDT)')
            return

        watchdog_timeout = self.timeout
        log.info('Starting the watchdog timer (WDT) with timeout {}ms'.format(watchdog_timeout))

        from machine import WDT
        self.wdt = WDT(timeout=watchdog_timeout)

        # Feed Watchdog once.
        self.feed()

    def resume(self):
        """ 
        Resume the watchdog.
        """
        log.info('Resuming watchdog')
        self.suspended = False
        self.reconfigure_minimum_timeout(self.timeout)

    def suspend(self):
        """
        Disabling a started watchdog is not possible, so let's
        just configure it to an ultra large timeout value.
        """

        log.info('Suspending watchdog')
        self.suspended = True
        self.reconfigure_minimum_timeout(999999999)

    def feed(self):
        """ 
        Feed the watchdog.
        """

        if not self.enabled:
            return

        # When in maintenance mode, try to suspend the watchdog.
        #if self.device.status.maintenance and not self.suspended:
        #    self.stop()

        # Always reconfigure to regular timeout when not in maintenance mode.
        #else:
        #    self.suspended = False
        #    self.wdt.init(self.timeout)

        # Feed the watchdog.
        log.debug('Feeding Watchdog')
        self.wdt.feed()

    def reconfigure_minimum_timeout(self, timeout):
        """

        :param timeout: 

        """
        if not self.enabled:
            return
        if timeout >= self.timeout:
            log.info('Reconfiguring watchdog timeout to {} milliseconds'.format(timeout))
            self.wdt.init(timeout)

    def adjust_for_interval(self, interval):
        """

        :param interval: 

        """
        if self.enabled and not self.suspended:
            watchdog_timeout = self.timeout / 1000.0
            if watchdog_timeout - 2 < interval:
                watchdog_timeout_effective = int((interval + 20) * 1000)
                log.warning('Reconfiguring original watchdog timeout {} as it is '
                            'smaller or near the configured sleep time {}'.format(watchdog_timeout, interval))
                self.reconfigure_minimum_timeout(watchdog_timeout_effective)



# -*- coding: utf-8 -*-
# (c) 2020 Richard Pobering <richard@hiveeyes.org>
# (c) 2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.util import get_platform_info

log = logging.getLogger(__name__)


class GPRSManager:
    """ """

    def __init__(self, manager, settings):
        self.manager = manager
        self.settings = settings

        # GPRS settings.
        self.gprs_settings = self.settings.get('networking.gprs')

        # Modem handle.
        self.modem = None

    def start(self):
        """ """
        log.info('[GPRS] Starting GPRS Manager')
        self.start_modem()
        self.connect()

    def start_modem(self):
        """ """

        # Resolve modem driver.
        if self.gprs_settings['driver'] != 'pythings-sim800':
            raise KeyError('Unable to setup unknown modem driver {}'.format(self.gprs_settings['driver']))

        from pythings_sim800 import Modem

        platform_info = get_platform_info()

        # Define two pins for clock and data.
        if platform_info.vendor == platform_info.MICROPYTHON.Vanilla:
            self.modem = Modem(
                MODEM_PWKEY_PIN=int(self.gprs_settings['pin_pwrkey'][1:]),
                MODEM_RST_PIN=int(self.gprs_settings['pin_reset'][1:]),
                MODEM_POWER_ON_PIN=int(self.gprs_settings['pin_power'][1:]),
                MODEM_TX_PIN=int(self.gprs_settings['pin_txd'][1:]),
                MODEM_RX_PIN=int(self.gprs_settings['pin_rxd'][1:]))

        elif platform_info.vendor == platform_info.MICROPYTHON.Pycom:
            self.modem = Modem(
                MODEM_PWKEY_PIN=self.gprs_settings['pin_pwrkey'],
                MODEM_RST_PIN=self.gprs_settings['pin_reset'],
                MODEM_POWER_ON_PIN=self.gprs_settings['pin_power'],
                MODEM_TX_PIN=self.gprs_settings['pin_txd'],
                MODEM_RX_PIN=self.gprs_settings['pin_rxd'])

        else:
            raise NotImplementedError('GPRS modem support is not implemented on this platform')

        # Initialize the modem.
        self.modem.initialize()

    def connect(self):
        # Connect the modem.
        self.modem.connect(apn=self.gprs_settings['apn'])

    def disconnect(self):
        # Disconnect Modem.
        self.modem.disconnect()

    def send_request(self, url, payload, content_type):
        response = self.modem.http_request(url, 'POST', payload, content_type)
        log.info('[GPRS] HTTP response status code: %s', response.status_code)
        log.info('[GPRS] HTTP response content: %s', response.content)
        return response



# -*- coding: utf-8 -*-
# (c) 2017-2018 David Gerber <https://github.com/geda>
# HX711 library for the LoPy.
# https://github.com/geda/hx711-lopy
"""
Updates
=======
(c) 2019 Ralf Lindlein <https://github.com/walterheisenberg>
License: GNU General Public License, Version 3
https://github.com/walterheisenberg/hivewatch_esp32/blob/master/lib/hx711lib.py

Setup
=====
::

    sc = HX711(0, 2)
    sc.set_scale(11.026667)
    sc.set_offset(130800.0)

    # Offset without wooden board
    #sc.set_offset(130748.3)

Notes
=====
- Some infos about low-pass filters
  https://microsoft.public.de.excel.narkive.com/PNakkyeu/tiefpassfilter

- The **time constant** means:

    - Lowers value will yield smoother low-pass filtering.
    - Use 0.9 to get more responsive values, 0.1 (default) means more filtered.

"""

from terkin.lib.hx711 import HX711
from terkin import logging

log = logging.getLogger(__name__)


class HX711Heisenberg(HX711):
    """ """

    def __init__(self, dout, pd_sck, gain=128):
        super().__init__(dout, pd_sck, gain)

        self.time_constant = 0.9

    def read_median(self, times=10):
        """Acquire multiple readings and return median.

        :param times:  (Default value = 10)

        """
        lst = []
        # Fixme: Think about feeding the watchdog here.
        for i in range(times):
            lst.append(self.read())
        sortedLst = sorted(lst)
        #print(sortedLst)
        lstLen = len(lst)
        index = (lstLen - 1) // 2

        if (lstLen % 2):
            return sortedLst[index]
        else:
            return (sortedLst[index] + sortedLst[index + 1])/2.0

    def get_avgkg(self, times=3):
        """

        :param times:  (Default value = 3)

        """
        return round(self.get_value(times) / self.SCALE, 3)

    def get_lpkg(self):
        """ """
        self.filtered += self.time_constant * (self.read() - self.filtered)
        return round((self.filtered - self.OFFSET) / self.SCALE, 3)

    def get_reading(self):
        """ """
        value = self.read_median()
        reading = WeightReading(value, self.OFFSET, self.SCALE)
        return reading

    def tare(self, times=15):
        """

        :param times:  (Default value = 15)

        """
        sum = self.read_average(times)
        self.set_offset(sum)
        log.debug('OFFSET = ' + str(sum))

    def set_scale(self, scale=None):
        """

        :param scale:  (Default value = None)

        """
        if scale is None:
            return self.SCALE
        self.SCALE = scale

    def set_offset(self, offset=None):
        """

        :param offset:  (Default value = None)

        """
        if offset is None:
            return self.OFFSET
        self.OFFSET = offset


class WeightReading:
    """ """

    def __init__(self, rawvalue, offset, scale):
        self.raw = rawvalue
        self.offset = offset
        self.scale = scale
        #self.raw_short = None
        self.kg = None
        self.compute()

    def compute(self):
        """ """
        #self.raw_short = self.raw
        try:
            self.kg = round((self.raw - self.offset) / self.scale, 3)
        except Exception as ex:
            log.exc(ex, 'Computing kg value failed')

    def get_data(self):
        """ """
        return self.__dict__



# -*- coding: utf-8 -*-
#
# Terkin Datalogger
# https://github.com/hiveeyes/terkin-datalogger
#
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019-2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2019-2020 Richard Pobering <richard@hiveeyes.org>
#
# License: GNU General Public License, Version 3
#
"""
Please check https://community.hiveeyes.org/t/operate-the-terkin-datalogger-sandbox/2332
in order to get an idea about how to operate the sandbox of this software.

Have fun!
"""
import os
import click
import logging
from terkin_cpython.util import setup_logging, configure_pythonpath, patch_system, start_bootloader, load_settings

# Configure logging.
setup_logging()
log = logging.getLogger()


@click.command()
@click.option("--config", required=True, help="Path to settings file.")
@click.option("--site-packages", help="Path to additional Python modules.")
@click.option("--daemon", is_flag=True, default=False, help="Run in daemon mode.")
@click.option("--ui", is_flag=True, default=False, help="Start user interface.")
def cli(**options):

    # Reconfigure module search path.
    if 'site_packages' in options and options['site_packages']:
        search_path = options['site_packages']
    else:
        search_path = os.path.abspath(os.path.join('.', 'dist-packages'))
    configure_pythonpath(search_path)

    # Bootstrap datalogger application.
    app = TerkinApplication(config=options['config'], daemon=options['daemon'])
    if options['ui']:
        app.start_ui()
    else:
        app.start()


class TerkinApplication:
    """Start the data logger application."""

    def __init__(self, config=None, daemon=False):
        self.config = config
        self.daemon = daemon
        self.bootloader = None
        self.datalogger = None

    def start(self):

        # Make environment compatible with CPython.
        patch_system()

        log.info('Starting bootloader')
        self.bootloader = start_bootloader()

        log.info('Loading modules')
        from terkin.datalogger import TerkinDatalogger

        settings = load_settings(settings_file=self.config)

        log.info('Setting up Terkin')
        self.datalogger = TerkinDatalogger(settings, platform_info=self.bootloader.platform_info)
        self.datalogger.setup()

        log.info('Starting Terkin')
        try:
            if self.daemon:
                self.datalogger.start()
            else:
                self.datalogger.duty_cycle()

        except KeyboardInterrupt:
            self.shutdown()

    def shutdown(self):

        log.info('Shutting down Terkin')

        if self.datalogger.device.status.networking:
            self.datalogger.device.networking.stop()

    def start_ui(self):

        # Make environment compatible with CPython.
        patch_system()

        from terkin.ui import TerkinUi

        ui = TerkinUi()
        try:
            ui.start()
            ui.example_listbox()
            #ui.example_menu()

        finally:
            ui.stop()



# -*- coding: utf-8 -*-
# (c) 2019-2020 Richard Pobering <richard@hiveeyes.org>
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import sys
import time
import socket
import machine
from terkin import logging
from terkin.util import format_exception, Eggtimer

log = logging.getLogger(__name__)


class NetworkManager:
    """ """

    def __init__(self, device, settings):
        self.device = device
        self.settings = settings

        self.device.watchdog.feed()

        self.wifi_manager = None
        self.lora_manager = None
        self.lte_manager = None
        self.gprs_manager = None

        self.mode_server = None
        self.http_api = None

    def stop(self):
        """ """
        self.stop_modeserver()

        # This helps the webserver to get rid of any listening sockets.
        # https://github.com/jczic/MicroWebSrv2/issues/8
        self.stop_httpserver()

        if self.wifi_manager:
            self.wifi_manager.stop()
            if self.device.status.maintenance is not True:
                self.wifi_manager.power_off()

        if self.lte_manager:
            self.lte_manager.stop()

    def start_wifi(self):
        """ """
        from terkin.network.wifi import WiFiManager
        self.wifi_manager = WiFiManager(manager=self, settings=self.settings)
        self.wifi_manager.start()

    def start_lora(self):
        """ """
        from terkin.network.lora import LoRaAdapter
        self.lora_manager = LoRaAdapter(network_manager=self, settings=self.settings)
        self.lora_manager.start()

    def start_lte(self):
        """ """
        from terkin.network.lte import SequansLTE
        self.lte_manager = SequansLTE(network_manager=self, settings=self.settings)
        self.lte_manager.start()

    def start_gprs(self):
        from terkin.network.gprs import GPRSManager
        self.gprs_manager = GPRSManager(manager=self, settings=self.settings)
        self.gprs_manager.start()

    def wait_for_ip_stack(self, timeout=5):
        """

        :param timeout:  (Default value = 5)

        """

        eggtimer = Eggtimer(duration=timeout)

        log.info('Waiting for the network stack to come up within %s seconds', timeout)
        while not eggtimer.expired():

            self.device.watchdog.feed()

            try:
                socket.getaddrinfo("localhost", 333)
                log.info('Network stack ready')
                return True

            except OSError as ex:
                #log.warning('Network stack not available: %s', format_exception(ex))
                pass

            # Report about progress.
            sys.stderr.write('.')
            #sys.stderr.flush()

            # Save power while waiting.
            machine.idle()
            time.sleep(0.25)

        # TODO: Make WiFi-agnostic.
        raise NetworkUnavailable('Could not connect to WiFi network')

    def wait_for_nic(self, timeout=5):
        """

        :param timeout:  (Default value = 5)

        """

        eggtimer = Eggtimer(duration=timeout)

        log.info('Waiting for network interface')
        while not eggtimer.expired():

            self.device.watchdog.feed()

            try:
                # TODO: Make WiFi-agnostic.
                if self.wifi_manager.is_connected():
                    log.info('Network interface ready')
                    return True

            except OSError as ex:
                log.warning('Network interface not available: %s', format_exception(ex))

            # Report about progress.
            sys.stderr.write('.')
            #sys.stderr.flush()

            # Save power while waiting.
            machine.idle()
            time.sleep(0.25)

        # TODO: Make WiFi-agnostic.
        raise NetworkUnavailable('Could not connect to WiFi network')

    def start_services(self):
        """ """

        # Start UDP server for pulling device into maintenance mode.
        if self.settings.get('services.api.modeserver.enabled', False):
            try:
                self.start_modeserver()
            except Exception as ex:
                log.exc(ex, 'Starting mode server failed')

        # Start HTTP server
        if self.settings.get('services.api.http.enabled', False):
            try:
                self.start_httpserver()
            except Exception as ex:
                log.exc(ex, 'Starting HTTP server failed')

    def start_modeserver(self):
        """Start UDP server for pulling device into maintenance mode."""

        # UDP server settings.
        #ip = self.wifi_manager.get_ip_address()
        ip = '0.0.0.0'
        port = self.settings.get('services.api.modeserver.port', 666)

        log.info('Starting mode server on {}:{}'.format(ip, port))
        from terkin.api.udp import UdpServer
        self.mode_server = UdpServer(ip, port)
        self.mode_server.start(self.handle_modeserver)

    def stop_modeserver(self):
        if self.mode_server:
            self.mode_server.stop()

    def start_httpserver(self):
        """Start HTTP server for managing the device."""
        from terkin.api.http import TerkinHttpApi
        storage = self.device.application_info.application.storage

        log.info('Setting up HTTP API')
        self.http_api = TerkinHttpApi(device=self.device, settings=self.settings, storage=storage)
        self.http_api.start()

    def stop_httpserver(self):
        if self.http_api:
            try:
                log.info('Shutting down HTTP server')
                self.http_api.stop()
            except Exception as ex:
                log.exc(ex, 'Shutting down HTTP server failed')

    def handle_modeserver(self, data, addr):
        """

        :param data: 
        :param addr: 

        """

        message = data.decode()

        if message == 'maintenance.enable()' and not self.device.status.maintenance:
            log.info('Enabling maintenance mode')
            self.device.status.maintenance = True
            self.device.watchdog.suspend()

        elif message == 'maintenance.disable()' and self.device.status.maintenance:
            log.info('Releasing maintenance mode')
            self.device.status.maintenance = False
            self.device.watchdog.resume()


class NetworkUnavailable(Exception):
    """ """
    pass



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import os
import time
import socket
import machine
import network


class MiniNet:
    """ 
    Used by the sandbox for 'make connect-wifi' to upload the program via ftp.
    Not used in the rest of the program.
    """

    def __init__(self):
        print('INFO:  Starting MiniNet')
        self.reconfigure_watchdog()
        self.station = network.WLAN()

    def reconfigure_watchdog(self, timeout_seconds=600):
        """

        :param timeout_seconds:  (Default value = 600)

        """
        try:
            from machine import WDT
            watchdog_timeout = timeout_seconds
            watchdog_timeout_effective = watchdog_timeout * 1000
            wdt = WDT(timeout=watchdog_timeout_effective)
            wdt.init(watchdog_timeout_effective)
            print('INFO:  Reconfigured watchdog to {} seconds'.format(watchdog_timeout))
        except:
            pass

    def activate_wifi_ap(self):
        """https://docs.pycom.io/firmwareapi/pycom/network/wlan/"""

        #self.station = network.WLAN(mode=network.WLAN.STA)
        try:
            self.station.disconnect()
        except:
            pass

        # Connect to WLAN
        print('INFO:  WiFi AP:  Starting access point')
        self.station.mode(network.WLAN.AP)

    def connect_wifi_sta(self, ssid, password, timeout=10000):
        """https://docs.pycom.io/firmwareapi/pycom/network/wlan/
        https://mike632t.wordpress.com/2017/04/11/connecting-my-wipy-to-my-wifi/

        :param ssid: 
        :param password: 
        :param timeout:  (Default value = 10000)

        """

        # Duplicate output on UART
        #uart = machine.UART(0, 115200)
        #os.dupterm(uart)

        # Disable telnet and FTP server before connecting to the network
        #server = network.Server()
        #server.deinit()

        # Connect to WLAN
        print('INFO:  WiFi STA: Scanning for networks')
        #self.station = network.WLAN(mode=network.WLAN.STA)
        try:
            self.station.disconnect()
        except:
            pass

        self.station.mode(network.WLAN.STA)
        nets = self.station.scan()
        ssids = [net.ssid for net in nets]

        # TODO: Sort networks by RSSI.
        #networks = [{'ssid': net.ssid, 'rssi': net.rssi} for net in nets]

        print('INFO:  WiFi STA: Networks found {}'.format(ssids))

        for net in nets:
            if net.ssid == ssid:
                # Try to connect twice.
                if self.connect_wifi_sta_single(ssid, net.sec, password, timeout=timeout):
                    break
                if self.connect_wifi_sta_single(ssid, net.sec, password, timeout=timeout):
                    break

        # Enable telnet and FTP server with new settings
        #server.init(login=('<user>', '<password>'), timeout=600)

        # Wait 10 seconds before continuing
        #time.sleep(10)

        #while not wlan.isconnected():
        #    time.sleep_ms(50)

        #self.wait_for_nic()

        if self.station.isconnected():
            print('INFO:  WiFi STA: Network configuration is {}'.format(self.station.ifconfig()))
            print('INFO:  Ready.')
            print()
            print('INFO:  WiFi STA: IP address is {}'.format(self.get_ip_address()))

        print()
        print('Note: Press CTRL+X or Ctrl+] to detach from the REPL')

    def connect_wifi_sta_single(self, ssid, authmode, password, timeout=10000):
        """

        :param ssid: 
        :param authmode: 
        :param password: 
        :param timeout:  (Default value = 10000)

        """

        print('INFO:  WiFi STA: Connecting to "{}"'.format(ssid))
        self.station.connect(ssid, auth=(authmode, password), timeout=timeout)

        try:
            # FIXME: This is a candidate for an infinite loop.
            while not self.station.isconnected():
                # Save power while waiting
                machine.idle()
                time.sleep_ms(250)

            print('INFO:  WiFi STA: Connected to "{}"'.format(ssid))

            return True

        except Exception as ex:
            print('ERROR: WiFi STA: Connecting to "{}" failed. Please check SSID and PASSWORD.\n{}'.format(ssid, ex))

    def wait_for_nic(self, retries=5):
        """

        :param retries:  (Default value = 5)

        """
        attempts = 0
        while attempts < retries:
            try:
                socket.getaddrinfo("localhost", 333)
                break
            except OSError as ex:
                print('Network interface not available: {}'.format(ex))
            print('Waiting for network interface')
            # Save power while waiting.
            machine.idle()
            time.sleep(0.25)
            attempts += 1
        print('Network interface ready')

    def get_ip_address(self):
        """ """
        return self.station.ifconfig()[0]



# -*- coding: utf-8 -*-
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor
from DS3231tokei import DS3231
from machine import RTC

log = logging.getLogger(__name__)


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create DS3231 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = DS3231Sensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)

    return sensor_object


class DS3231Sensor(AbstractSensor):
    """
    A generic DS3231 sensor/actor component.
    The DS3231 has a RTC which can be read & set. It also has a good temperature sensor.
    If the battery on the RTC has to be changed the time is set to 1.1.2000, 0.00h
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory.
    """

    def __init__(self, settings=None):

        # from terkin/sensor/core.py class AbstractSensor
        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.address = 0x68

    def start(self):
        """ Getting the bus """
        if self.bus is None:
            raise KeyError("Bus missing for DS3231")

        # Initialize the hardware driver.
        try:
            self.driver = DS3231(i2c=self.bus.adapter)
            return True

        except Exception as ex:
            log.exc(ex, 'DS3231 hardware driver failed')

        # set date/time of RTC
        self.set_time()

    def read(self):
        """ the DS3231 has a sensor for temperature compensation """

        if self.bus is None or self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        #log.info('Acquire reading from DS3231')

        data = {}

        temperature = self.driver.getTemperature()

        # Prepare readings.
        values = {
            "temperature": temperature,
        }

        # Build telemetry payload.
        fieldnames = values.keys()
        for name in fieldnames:
            fieldname = self.format_fieldname(name, hex(self.address))
            value = values[name]
            data[fieldname] = value

        if not data:
            log.warning("I2C device {} has no value: {}".format(self.address, data))

        log.debug("I2C data:     {}".format(data))

        return data

    def set_time(self):
        """ set the system time to RTC time """

        rtc = RTC()
        [year,month,day,dotw,hour,minute,second] = self.driver.getDateTime() # get the date/time from the DS3231
        if year > 2019: # check valid data 
            rtc.init((year,month,day,dotw,hour,minute,second,0))    # set date/time
            log.debug("Time set:     {}".format(rtc.datetime()))
        else:
            log.warning("DS3231 date/time not set, not setting RTC")



    


# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create ADS1x15 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = ADS1x15Sensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)

    return sensor_object


class ADS1x15Sensor(AbstractSensor):
    """
    A generic ADS1x15 sensor component.
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory.
    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        self.handler = None
        self.driver = None
        self.channel = self.settings['channel']
        self.model = self.settings['model']

        # Can be overwritten by ``.set_address()``.
        self.address = self.settings.get('address', 0x48)

    def start(self):

        if self.bus is None:
            raise KeyError("Bus missing for ADS1x15 sensor")

        # Initialize the hardware driver.
        try:

            # MicroPython
            if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
                # TODO: Integrate https://github.com/robert-hh/ads1x15 here.
                raise NotImplementedError('ADS1x15 driver not implemented on MicroPython')

            # CPython
            elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:

                # Select model.
                if self.model == 'ads1115':
                    import adafruit_ads1x15.ads1115 as ADS
                    driver_class = ADS.ADS1115
                elif self.model == 'ads1015':
                    import adafruit_ads1x15.ads1015 as ADS
                    driver_class = ADS.ADS1015
                else:
                    log.warning('ADS Model not supported')
                # Setup driver.
                from adafruit_ads1x15.analog_in import AnalogIn
                self.handler = driver_class(i2c=self.bus.adapter, address=self.address)
                self.driver = AnalogIn(self.handler, getattr(ADS, self.channel))

            else:
                raise NotImplementedError('ADS1x15 driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'ADS1x15 hardware driver failed')

    def read(self):

        if self.bus is None or self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        # MicroPython
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
            raise NotImplementedError('ADS1x15 driver not implemented on MicroPython')

        # CPython
        elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
            values = {
                # "value_raw": self.driver.value,
                # "voltage_raw": self.driver.voltage,
                "voltage": self.driver.voltage * self.settings['multiplicator'],
            }

        # Build telemetry payload.
        data = {}
        for key, value in values.items():
            name = '{}.{}'.format(key, self.model)
            fieldname = self.format_fieldname(name, hex(self.address), self.channel.lower())
            data[fieldname] = value

        return data



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import utime
import machine
import ubinascii
from network import Bluetooth
from terkin.network.ble.util import BluetoothUuid, sint16, uint16, encode_temperature_2a1c


class BluetoothApi:
    """Many examples integrated from
    https://microchipdeveloper.com/wireless:ble-gatt-data-organization
    https://development.pycom.io/firmwareapi/pycom/network/bluetooth/
    https://github.com/sandeepmistry/arduino-BLEPeripheral
    https://github.com/kriswiner/nRF52832DevBoard/blob/master/BMP280_nRF52.ino
    
    https://forum.pycom.io/topic/2826/how-to-send-data-to-characteristic-over-ble/17


    """

    def __init__(self):
        self.bt = None

    def start(self):
        """ """

        print('Starting Bluetooth')
        self.bt = Bluetooth()

        # Default
        #self.bt.init(id=0, mode=Bluetooth.BLE, antenna=Bluetooth.INT_ANT, modem_sleep=True)
        #self.bt.init(id=0, antenna=Bluetooth.INT_ANT, modem_sleep=False)
        self.bt.init(modem_sleep=False)

        return

        print('Entering main loop')
        while True:
            print('--- loop ---')

            adv = self.bt.get_adv()
            print('adv:', adv)

            # Give the system some breath.
            print('machine.idle()')
            machine.idle()

            utime.sleep(10.0)
            continue

    def scan(self, duration):
        """

        :param duration: 

        """
        duration = int(duration)
        print('Starting Bluetooth scan for {} seconds'.format(duration))
        self.bt.start_scan(duration)
        while self.bt.isscanning():
            #print('is-scanning')
            adv = self.bt.get_adv()
            if adv:
                # try to get the complete name
                name_short = self.bt.resolve_adv_data(adv.data, Bluetooth.ADV_NAME_SHORT)
                name_complete = self.bt.resolve_adv_data(adv.data, Bluetooth.ADV_NAME_CMPL)
                print('names:', name_short, name_complete)

                service_data = self.bt.resolve_adv_data(adv.data, Bluetooth.ADV_SERVICE_DATA)
                print('service_data:', service_data)

                mfg_data = self.bt.resolve_adv_data(adv.data, Bluetooth.ADV_MANUFACTURER_DATA)
                if mfg_data:
                    # try to get the manufacturer data (Apple's iBeacon data is sent here)
                    data = ubinascii.hexlify(mfg_data)
                    print('data:', data)

        print('Bluetooth scanning stopped')

    def find_heart_rate(self):
        """ """
        adv = self.bt.get_adv()
        if adv and self.bt.resolve_adv_data(adv.data, Bluetooth.ADV_NAME_CMPL) == 'Heart Rate':
            try:
                conn = self.bt.connect(adv.mac, timeout=5000)
                services = conn.services()
                for service in services:
                    utime.sleep(0.050)
                    if type(service.uuid()) == bytes:
                        print('Reading chars from service = {}'.format(service.uuid()))
                    else:
                        print('Reading chars from service = %x' % service.uuid())
                    chars = service.characteristics()
                    for char in chars:
                        if (char.properties() & Bluetooth.PROP_READ):
                            print('char {} value = {}'.format(char.uuid(), char.read()))
                conn.disconnect()
            except:
                print("Error while connecting or reading from the BLE device")
        else:
            utime.sleep(0.050)

        utime.sleep(10.0)

    def advertise(self):
        """https://development.pycom.io/firmwareapi/pycom/network/bluetooth/"""

        device_name = get_device_name()
        print('Advertising device {}'.format(device_name))
        # TODO: Get from settings
        self.bt.set_advertisement(name=device_name, service_uuid=b'1234567890123456')

        def conn_cb(bt_o):
            """

            :param bt_o: 

            """
            print("Callback happened")
            # Returns flags and clear internal registry
            events = bt_o.events()
            if events & Bluetooth.CLIENT_CONNECTED:
                print("Client connected:", events)
            elif events & Bluetooth.CLIENT_DISCONNECTED:
                print("Client disconnected")

        self.bt.callback(trigger=Bluetooth.CLIENT_CONNECTED | Bluetooth.CLIENT_DISCONNECTED, handler=conn_cb)

        self.bt.advertise(True)

    def start_service(self):
        """https://development.pycom.io/firmwareapi/pycom/network/bluetooth/gattsservice/
        https://development.pycom.io/firmwareapi/pycom/network/bluetooth/gattscharacteristic/
        
        https://www.bluetooth.com/specifications/gatt/characteristics/
        https://docs.python.org/3/library/struct.html
        
        The format field determines how a single value contained in the Characteristic Value is formatted.
        The information contained on this page is referenced in Bluetooth® Core Specification Volume 3, Part G, Section 3.3.3.5.2.
        https://www.bluetooth.com/specifications/assigned-numbers/format-types/


        """
        print('Starting service')

        # A. Battery Service
        # UUID: 180F
        # Abstract: The Battery Service exposes the state of a battery within a device
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.service.battery_service.xml
        battery = self.bt.service(uuid=0x180F, isprimary=True, nbr_chars=1, start=True)

        # Battery Level
        # UUID: 2A19
        # Abstract: The current charge level of a battery.
        #           100% represents fully charged while 0% represents fully discharged
        # Format: uint8
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.characteristic.battery_level.xml
        battery_level = battery.characteristic(uuid=0x2A19, value=78)

        # B. Environmental Sensing Service (ESS)
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.service.environmental_sensing.xml
        # https://www.bluetooth.com/specifications/assigned-numbers/environmental-sensing-service-characteristics/
        # https://github.com/sandeepmistry/arduino-BLEPeripheral/issues/215
        # https://github.com/sandeepmistry/arduino-nRF5/issues/248
        environment = self.bt.service(uuid=0x181A, isprimary=True, nbr_chars=5, start=True)

        # Temperature
        # UUID: 2A6E
        # InformativeText: Unit is in degrees Celsius with a resolution of 0.01 degrees Celsius
        # Format: sint16
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.characteristic.temperature.xml
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.characteristic.temperature_measurement.xml
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.characteristic.temperature_type.xml
        # https://github.com/ARMmbed/ble/blob/master/ble/services/HealthThermometerService.h
        # https://github.com/ARMmbed/ble/blob/master/ble/services/EnvironmentalService.h

        # Temperature Celsius
        # UUID: 2A1F
        # Format: sint16
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.temperature_celsius.xml
        self.temp = 42.42
        temperature = environment.characteristic(uuid=0x2A1F)
        def temp_getter(chr):
            """

            :param chr: 

            """
            print('Getting characteristic:', chr)
            value = sint16(self.temp * 10)
            print('Value:', value)
            self.temp += 1
            return value
        print('Adding characteristic callback')
        temperature_callback = temperature.callback(trigger=Bluetooth.CHAR_READ_EVENT, handler=temp_getter)

        # Temperature Measurement
        # UUID: 2A1C
        # Format: Variable length structure
        # Value Format: FLOAT (IEEE-11073 32-bit FLOAT)
        # Abstract: The Temperature Measurement characteristic is a variable length structure
        # containing a Flags field, a Temperature Measurement Value field and, based upon the
        # contents of the Flags field, optionally a Time Stamp field and/or a Temperature Type field.
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.temperature_measurement.xml
        print('Adding Temperature Measurement')
        temperature_measurement = environment.characteristic(uuid=0x2A1C, value=encode_temperature_2a1c(42.428423))

        # Scientific Temperature
        # UUID: 2A3C
        # Format: float64 (IEEE-754 64-bit floating point)
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.scientific_temperature_celsius.xml
        # http://forum.espruino.com/conversations/306536/
        # https://stackoverflow.com/questions/36655172/how-to-read-a-ble-characteristic-float-in-swift

        # For the 'f', 'd' and 'e' conversion codes, the packed representation uses the IEEE 754
        # binary32, binary64 or binary16 format (for 'f', 'd' or 'e' respectively), regardless of
        # the floating-point format used by the platform.
        # https://docs.python.org/3/library/struct.html
        #print('Adding Scientific Measurement')
        #temperature_sci = environment.characteristic(uuid=0x2A3C, value=float64(42.42))

        # Humidity
        # UUID: 2A6F
        # InformativeText: Unit is in percent with a resolution of 0.01 percent
        # Format: uint16
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.characteristic.humidity.xml
        humidity1 = environment.characteristic(uuid=0x2A6F, value=uint16(86.86 * 100))
        humidity2 = environment.characteristic(uuid=0x2A6F, value=uint16(50.55 * 100))

        # Pressure
        # UUID: 2A6D
        # InformativeText: Unit is in pascals with a resolution of 0.1 Pa
        # Format: uint32
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.pressure.xml

        # Weight
        # UUID: 2A98
        # InformativeText: Unit is in kilograms with a resolution of 0.005
        # Format: uint16
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.weight.xml
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.weight_measurement.xml
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.weight_scale_feature.xml
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.characteristic.weight.xml
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.characteristic.weight_measurement.xml
        # https://github.com/oesmith/gatt-xml/blob/master/org.bluetooth.characteristic.weight_scale_feature.xml
        weight = environment.characteristic(uuid=0x2A98, value=uint16(25.44 * 2 * 100))

        # GPS
        # https://forum.pycom.io/topic/5012/can-only-see-value-of-first-gatt-characteristic
        """
        lan = bt.service(0x1819, nbr_chars=3)
        # 0.001 precision
        lan.characteristic(uuid=0x2AAE, value=int(gps.lat * 1000))
        lan.characteristic(uuid=0x2AAF, value=int(gps.lon * 1000))
        lan.characteristic(uuid=0x2AB3, value=int(gps.alt * 1000))
        """

        # Generic Access
        # UUID: 1800
        # Abstract: The generic_access service contains generic information
        # about the device. All available Characteristics are readonly.
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Services/org.bluetooth.service.generic_access.xml

        # Generic Attribute Service
        # UUID: 1801
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Services/org.bluetooth.service.generic_attribute.xml
        # String Characteristic
        # UUID: 2A3D
        # A generic UTF8 string which may be used in Services requiring strings.
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Characteristics/org.bluetooth.characteristic.string.xml
        srv1 = self.bt.service(uuid=BluetoothUuid.to_bytes_le('15ECCA29-0B6E-40B3-9181-BE9509B53200'), isprimary=True, nbr_chars=2, start=True)
        #generic_attribute = self.bt.service(uuid=0x1800, isprimary=True, nbr_chars=2, start=True)
        #wifi_ssid = generic_attribute.characteristic(uuid=0x2A3D, properties=Bluetooth.PROP_BROADCAST | Bluetooth.PROP_INDICATE | Bluetooth.PROP_NOTIFY | Bluetooth.PROP_READ | Bluetooth.PROP_WRITE, value="default")
        #wifi_ssid = generic_attribute.characteristic(uuid=0x2A3D, properties=Bluetooth.PROP_NOTIFY | Bluetooth.PROP_READ | Bluetooth.PROP_WRITE, value="default")
        wifi_ssid = srv1.characteristic(uuid=0x2A3D)
        #wifi_password = generic_attribute.characteristic(uuid=0x2A3D, properties=Bluetooth.PROP_READ | Bluetooth.PROP_WRITE)
        #print('wifi_ssid:', wifi_ssid, dir(wifi_ssid))
        #cfg = wifi_ssid.config()
        #print('cfg:', cfg)
        #wifi_ssid.config(name='hallodri')

        print('wifi_ssid:', wifi_ssid)
        print('wifi_ssid:', dir(wifi_ssid))

        def wifi_callback(characteristic):
            """

            :param characteristic: 

            """
            print('Characteristic callback:', characteristic)
            #print('attr_obj:', characteristic.attr_obj)
            events = characteristic.events()
            if events & Bluetooth.CHAR_WRITE_EVENT:
                print("Write request with value = {}".format(characteristic.value()))
            else:
                pass
            value = characteristic.value()
            print('Reading characteristic:', value)

        print('Adding characteristic callback')
        wifi_ssid_callback = wifi_ssid.callback(trigger=Bluetooth.CHAR_READ_EVENT | Bluetooth.CHAR_WRITE_EVENT, handler=wifi_callback)
        #wifi_password_callback = wifi_password.callback(trigger=Bluetooth.CHAR_READ_EVENT | Bluetooth.CHAR_WRITE_EVENT, handler=wifi_callback)

        #print('instance:', wifi_ssid.instance())

        # Characteristic User Description
        # UUID: 0x2901,0x2902
        print(wifi_ssid_callback)
        print(dir(wifi_ssid_callback))
        srv1.characteristic(uuid=0x2901, value="hello world 1")
        #wifi_ssid_callback.characteristic(uuid=0x2901, value="hello world 2")

        # Automation IO
        # UUID: 1815
        # Abstract: The Automation IO service is used to expose the analog
        # inputs/outputs and digital input/outputs of a generic IO module (IOM).
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Services/org.bluetooth.service.automation_io.xml

        # User Data
        # UUID: 181C
        # Abstract: This service exposes user-related data in the sports and fitness environment.
        # This allows remote access and update of user data by a Client as well as the synchronization
        # of user data between a Server and a Client.
        # https://www.bluetooth.com/wp-content/uploads/Sitecore-Media-Library/Gatt/Xml/Services/org.bluetooth.service.user_data.xml
        #user_data = self.bt.service(uuid=0x0123456789ABCDEF, isprimary=True, nbr_chars=1, start=True)


def get_device_id():
    """ """
    import machine
    from ubinascii import hexlify
    return hexlify(machine.unique_id()).decode()


def get_device_name():
    """ """
    import os
    osname = os.uname().sysname
    sysid = get_device_id()
    name = 'Terkin-{}-{}'.format(osname, sysid)
    return name


if __name__ == '__main__':
    """Main application entrypoint."""
    global btapi
    btapi = BluetoothApi()
    btapi.start()
    #btapi.scan(2)
    btapi.advertise()
    btapi.start_service()

    #print('Sleeping for 20 seconds')
    #utime.sleep(20)



# -*- coding: utf-8 -*-
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019-2020 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3
import time
from terkin import logging
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create DS18x20 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = DS18x20Sensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)
    return sensor_object


class DS18x20Sensor(AbstractSensor):
    """
    A generic DS18B20 sensor component.
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory
    """

    def start(self):
        """
        Setup the DS18x20 sensor driver.
        :return:
        """

        # Ensure a bus object exists and is ready.
        self.ensure_bus()

        # Initialize the DS18x20 hardware driver.
        onewire_bus = self.bus.adapter
        try:

            use_native_driver = self.bus.settings.get('driver') == 'native'

            platform_info = get_platform_info()

            # Vanilla MicroPython >=1.11
            if platform_info.vendor == platform_info.MICROPYTHON.Vanilla or use_native_driver:
                import ds18x20_native
                self.driver = ds18x20_native.DS18X20(onewire_bus)

            # Pycom MicroPython 1.9.4
            elif platform_info.vendor == platform_info.MICROPYTHON.Pycom:
                import ds18x20_python
                self.driver = ds18x20_python.DS18X20(onewire_bus)

            # RaspberryPi
            elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
                from terkin.sensor.linux import LinuxSysfsDS18B20
                self.driver = LinuxSysfsDS18B20(onewire_bus)

            else:
                raise NotImplementedError('DS18x20 driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'DS18x20 hardware driver failed')
            return False

    def read(self):
        """
        - Start conversion on all DS18x20 sensors.
        - Wait for conversion.
        - Read all DS18x20 sensors.
        """

        if self.bus is None or self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        log.info('Acquire readings from all DS18x20 sensors attached to bus "{}"'.format(self.bus.name))

        # Start conversion on all DS18x20 sensors.
        self.driver.convert_temp()
        time.sleep_ms(750)

        # Read scratch memory of each sensor.
        data = self.read_devices()

        if not data:
            log.warning('No data from any DS18x20 devices on bus "{}"'.format(self.bus.name))

        log.debug('Data from 1-Wire bus "{}" is "{}"'.format(self.bus.name, data))

        return data

    def get_effective_devices(self):
        """
        Return list of all DS18x20 devices.
        """

        log.info('Start conversion for DS18x20 devices on bus "{}"'.format(self.bus.name))
        effective_devices = []
        for device in self.bus.devices:

            address = self.bus.device_address_ascii(device)
            device_settings = self.get_device_settings(address)

            enabled = device_settings.get('enabled')
            if enabled is False:
                log.info('Skipping DS18x20 device "{}"'.format(address))
                continue

            effective_devices.append(device)

        return effective_devices

    def read_devices(self):
        """
        Read all DS18x20 sensors and return data dictionary.
        """

        data = {}
        devices = self.get_effective_devices()
        for device in devices:

            address = self.bus.device_address_ascii(device)
            log.info('Reading DS18x20 device "{}"'.format(address))
            try:
                value = self.driver.read_temp(device)
            except Exception as ex:
                log.exc(ex, "Reading DS18x20 device {} failed".format(address))
                continue

            # Evaluate device response.
            if value is not None:

                # TODO: Mask power-on reset value 0550h (85°) here.
                # https://github.com/micropython/micropython/pull/5338
                # Maybe it's not only 85°. Sometimes it's also 25°(!).
                # https://github.com/cpetrich/counterfeit_DS18B20

                try:
                    # Compute telemetry field name.
                    fieldname = self.format_fieldname('temperature', address)

                    # Apply value offset.
                    offset = self.get_setting(address, 'offset')
                    if offset is not None:
                        log.info('Adding offset {} to value {} from device "{}"'.format(offset, value, address))
                        value += offset

                    # Add value to telemetry message.
                    data[fieldname] = value

                except Exception as ex:
                    log.exc(ex, 'Processing data from DS18x20 device "{}" failed'.format(address))
                    continue

            else:
                log.warning('No response from DS18x20 device "{}"'.format(address))

        return data

    def get_setting(self, address, name, default=None):
        """

        :param address: 
        :param name: 
        :param default:  (Default value = None)

        """
        settings = self.get_device_settings(address)
        value = settings.get(name, default)
        return value

    def get_device_settings(self, address):
        """

        :param address: 

        """

        # Compute ASCII representation of device address.
        address = self.bus.device_address_ascii(address)

        # Get device-specific settings from configuration.
        for device_settings in self.settings.get('devices', []):
            if device_settings['address'].lower() == address:
                return device_settings

        return {}

    def get_device_description(self, address):
        """

        :param address: 

        """
        device_settings = self.get_device_settings(address)
        return device_settings.get('description')



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3

from .core import SensorManager, BusType, OneWireBus, I2CBus
from .common import AbstractSensor



# -*- coding: utf-8 -*-
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# (c) 2019-2020 Richard Pobering <richard@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.model import SensorReading
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create BME280 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = BME280Sensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)

    return sensor_object


class BME280Sensor(AbstractSensor):
    """
    A generic BME280 sensor component.
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory.
    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.address = self.settings.get('address', 0x76)

    def start(self):
        """
        Setup the BME280 sensor driver.

        :return:
        """

        # Ensure a bus object exists and is ready.
        self.ensure_bus()

        # Initialize the hardware driver.
        try:

            # MicroPython
            if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
                from bme280_float import BME280
                self.driver = BME280(address=self.address, i2c=self.bus.adapter)

            # Adafruit CircuitPython
            elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:
                import adafruit_bme280
                self.driver = adafruit_bme280.Adafruit_BME280_I2C(i2c=self.bus.adapter, address=self.address)

            # RPi.bme280
            elif platform_info.vendor == platform_info.MICROPYTHON.Odroid:
                import bme280
                self.calibration_params = bme280.load_calibration_params(self.bus.adapter, self.address)
                self.driver = bme280.sample(self.bus.adapter, self.address)

            else:
                raise NotImplementedError('BME280 driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'BME280 hardware driver failed')
            return False

    def read(self):
        """
        Read the BME280 sensor.

        :return: SensorReading
        """

        if self.bus is None or self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        #log.info('Acquire reading from BME280')

        data = {}

        # MicroPython
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:

            t, p, h = self.driver.read_compensated_data()

            # Prepare readings.
            values = {
                "temperature": t,
                "pressure": p / 100,
                "humidity": h,
            }

        # Adafruit CircuitPython
        elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:

            values = {
                "temperature": self.driver.temperature,
                "humidity": self.driver.humidity,
                "pressure": self.driver.pressure,
            }

        # Adafruit CircuitPython
        elif platform_info.vendor == platform_info.MICROPYTHON.Odroid:

            values = {
                "temperature": self.driver.temperature,
                "humidity": self.driver.humidity,
                "pressure": self.driver.pressure,
            }


        # Build telemetry payload.
        # TODO: Push this further into the telemetry domain.
        fieldnames = values.keys()
        for name in fieldnames:
            fieldname = self.format_fieldname(name, hex(self.address))
            value = values[name]
            data[fieldname] = value

        if not data:
            log.warning("I2C device {} has no value: {}".format(self.address, data))

        log.debug("I2C data:     {}".format(data))

        reading = SensorReading()
        reading.sensor = self
        reading.data = data

        return reading

    @staticmethod
    def int_to_float(t, p, h):
        """Obtains readings from BME280's (integer variant) ``read_compensated_data()`` method.
        
        
        Integer formats
        ===============
        
        Temperature
        -----------
        The temperature in hundredths of a degree Celsius.
        For example, the value 2534 indicates a temperature of 25.34 degrees.
        
        Pressure
        --------
        The atmospheric pressure. This 32-bit value consists of 24 bits
        indicating the integer value, and 8 bits indicating the fractional
        value. To get a value in Pascals, divide the return value by 256.
        For example, a value of 24674867 indicates 96386.2Pa, or 963.862hPa.
        
        Humidity
        --------
        The relative humidity. This 32-bit value consists of 22 bits
        indicating the integer value, and 10 bits indicating the fractional
        value. To get a value in %RH, divide the return  value by 1024.
        For example, a value of 47445 indicates 46.333%RH.
        
        -- https://github.com/robert-hh/BME280#integer-formats

        :param t: 
        :param p: 
        :param h: 

        """
        # Prepare readings.
        values = {}
        if t:
            values["temperature"] = t / 100

        if p:
            p = p // 256
            pi = p // 100
            pd = p - pi * 100
            values["pressure"] = float("{}.{:02d}".format(pi, pd))

        if h:
            hi = h // 1024
            hd = h * 100 // 1024 - hi * 100
            values["humidity"] = float("{}.{:02d}".format(hi, hd))

        return values



# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst'), encoding="utf-8").read()

requires = [

    # 3rd-party libraries
    'terkin-micropython-libraries==0.12.0',

    # Mocks some HAL modules not available on CPython.
    'mock==4.0.2',
    'esp32-machine-emulator==1.1.3',

    # Sensors
    'psutil<=5.8.0,<6',

    # Telemetry
    'requests',

    # Command line interface
    'click==7.1.2',

    # UART access.
    'pyserial==3.4',

    # GPSD client library.
    'gps==3.19',

    # Adafruit CircuitPython libraries.
    'Adafruit-Blinka>=5.4.0,<6',
    'adafruit-circuitpython-busdevice==4.3.1',
    'adafruit-circuitpython-bme280==2.4.1',
    'adafruit-circuitpython-ads1x15==2.2.1',
    'adafruit-circuitpython-si7021==3.2.1',
    'adafruit-circuitpython-bmp280==3.2.1',
    'adafruit-circuitpython-ina219==3.4.2',


    # Victron Energy VE.Direct text protocol driver.
    #'git+https://github.com/karioja/vedirect@f74c0f2',

]

extras = {
    'sbc': [

        # I2C SMBus HAL module.
        'smbus==1.1.post2',

        # SPI HAL module.
        'spidev==3.4',

        # RPI USV+ Raspberry Pi - USV+
        'rpi-piusv==0.1.0',

        # Raspberry Pi utilities.
        'gpiozero==1.5.1',
    ],
    'lorawan': [
        # Required for LoRaWAN.
        'pycrypto==2.6.1',
    ],
}

setup(name='terkin',
      version='0.12.0',
      description='A flexible data logger for MicroPython and CPython',
      long_description=README,
      license="AGPL 3, EUPL 1.2",
      classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: MicroPython",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Communications",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Archiving",
        "Topic :: System :: Networking :: Monitoring",
        "Topic :: Utilities",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS"
        ],
      author='Andreas Motl',
      author_email='andreas.motl@terkin.org',
      url='https://github.com/hiveeyes/terkin-datalogger',
      keywords='sensor networks data acquisition transformation daq routing '
               'telemetry m2m iot mqtt http',
      package_dir={
          '': 'src/lib',
      },
      packages=find_packages('src/lib'),
      py_modules=[
          'umal',
          'mininet',
      ],
      include_package_data=True,
      package_data={
      },
      zip_safe=False,
      test_suite='test',
      install_requires=requires,
      extras_require=extras,
      #tests_require=extras['test'],
      dependency_links=[
      ],
      entry_points={
          'console_scripts': [
              'terkin = terkin_cpython.main:cli',
          ],
      },
)



# -*- coding: utf-8 -*-
# (c) 2019 Richard Pobering <richard@hiveeyes.org>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
import time
import machine
import network
import binascii

from terkin import logging
from terkin.sensor.system import AbstractSystemSensor
from terkin.util import get_platform_info, format_mac_address, backoff_time, Stopwatch

log = logging.getLogger(__name__)


class WiFiManager:
    """
    Manages all things relevant to the WiFi connection.
    """

    def __init__(self, manager, settings):

        self.manager = manager
        self.settings = settings

        # WIFI settings.
        self.phy = self.settings.get('networking.wifi.phy', {})
        self.stations = self.settings.get('networking.wifi.stations')
        self.station = None

        # Stopwatch for keeping track of time.
        log.info('Starting stopwatch')
        self.stopwatch = Stopwatch()
        log.info('Started stopwatch successfully')

        # Helper for the ``stay_connected`` thread.
        self.is_running = False

        # Reference to platform information.
        self.platform_info = get_platform_info()

    def start(self):
        """ """

        self.start_interface()

        # Check WiFi connectivity.
        if not self.is_connected():

            # Free up some memory.
            # self.manager.device.run_gc()

            # Connect to WiFi network and wait for successful connection.
            self.connect_once()

        else:

            log.info("WiFi STA: Network connection already established, will skip scanning and resume connectivity.")
            self.print_short_status()

            # Give system some breath.
            #time.sleep(0.25)

            # Inform about networking status.
            #self.print_short_status()
            self.print_address_status()

        # Start thread to monitor WiFi connection and reconnect if required.
        try:
            import _thread
            self.is_running = True
            log.info('Starting WiFi connection monitor')
            _thread.start_new_thread(self.stay_connected_invoke, ())
        except:
            log.warning('Failed starting WiFi connection monitor')

    def stop(self):
        log.info('Shutting down WiFi connection monitor')
        self.is_running = False

    def start_interface(self):
        """
        Genuine MicroPython:
        https://docs.micropython.org/en/latest/library/network.WLAN.html
        
        Pycom MicroPython:
        https://docs.pycom.io/tutorials/all/wlan.html
        https://github.com/pycom/pydocs/blob/master/firmwareapi/pycom/network/wlan.md
        """

        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            self.station = network.WLAN()
        else:
            log.info('WiFi STA: Will exclusively use STA mode on this platform. AP mode not implemented yet.')
            self.station = network.WLAN(network.STA_IF)

        #if machine.reset_cause() == machine.SOFT_RESET:
        #   print("WiFi STA: Network connection after SOFT_RESET.")
        #    self.print_short_status()
        #    # Inform about networking status.
        #    self.print_address_status()
        #    return True

        # Save the default ssid and auth for restoring AP mode later
        #original_ssid = self.station.ssid()
        #original_auth = self.station.auth()

        # Inform about networking status.
        self.print_address_status()

        # Setup network interface.
        log.info("WiFi: Starting interface")

        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            self.configure_antenna()

            # FIXME: Make STA or AP or STA_AP configurable.
            self.station.mode(network.WLAN.STA_AP)

            # Initialize the WiFi peripheral.
            self.station.init()

        else:
            self.station.active(True)

    def configure_antenna(self):
        """ """
        # https://community.hiveeyes.org/t/signalstarke-des-wlan-reicht-nicht/2541/11
        # https://docs.pycom.io/firmwareapi/pycom/network/wlan/

        antenna_external = self.phy.get('antenna_external', False)

        if not hasattr(self.station, 'antenna'):
            # Select antenna, 0=chip, 1=external.
            antenna_value = 0
            if antenna_external:
                antenna_value = 1
            self.station.config(antenna=antenna_value)
            return

        if antenna_external:
            antenna_pin = self.phy.get('antenna_pin')
            log.info('WiFi: Using external antenna on pin %s', antenna_pin)

            # To use an external antenna, set P12 as output pin.
            from machine import Pin
            Pin(antenna_pin, mode=Pin.OUT)(True)

            # Configure external WiFi antenna.
            self.station.antenna(network.WLAN.EXT_ANT)

        else:
            log.info('WiFi: Using internal antenna')
            self.station.antenna(network.WLAN.INT_ANT)

    def enable_ap(self):
        """ """
        # Todo: Reenable WiFi AP mode in the context of an "initial configuration" mode.
        """
        log.info('WiFi: Switching to AP mode')
        # WLAN.AP, original_ssid, original_auth, WLAN.INT_ANT
        # TOOD: Make default channel configurable
        self.station.init(mode=WLAN.AP, ssid=original_ssid, auth=original_auth, channel=6, antenna=WLAN.INT_ANT)
        """
        pass

    def get_configured_stations(self):
        return frozenset([station['ssid'] for station in self.stations])

    def connect_once(self):
        """ """

        # Prepare information about known WiFi networks.
        networks_known = self.get_configured_stations()

        try:
            self.connect_stations(networks_known)

        except Exception as ex:
            log.exc(ex, 'WiFi STA: Connecting to configured networks "{}" failed'.format(list(networks_known)))

    def stay_connected_invoke(self):
        try:
            self.stay_connected()
        except KeyboardInterrupt:
            log.info("Received KeyboardInterrupt within WiFi connection monitor")
            self.stop()
            raise

    def stay_connected(self):
        """ """

        # Prepare information about known WiFi networks.
        networks_known = self.get_configured_stations()

        # Attempt to connect to known/configured networks.
        attempt = 0
        while self.is_running:

            delay = 1

            if self.is_connected():
                attempt = 0

            else:
                log.info("WiFi STA: Connecting to configured networks: %s. "
                         "Attempt: #%s", list(networks_known), attempt + 1)
                try:
                    self.connect_stations(networks_known)

                except KeyboardInterrupt:
                    raise

                except Exception as ex:
                    log.exc(ex, 'WiFi STA: Connecting to configured networks "{}" failed'.format(list(networks_known)))
                    delay = backoff_time(attempt, minimum=1, maximum=600)
                    log.info('WiFi STA: Retrying in {} seconds'.format(delay))

                attempt += 1

            machine.idle()
            time.sleep(delay)

    def is_connected(self):
        """
        Check if connected to WiFi access point and
        connection yielded a valid IP address.
        """
        try:
            # ``isconnected()`` returns True when connected to a
            # WiFi access point *and* having a valid IP address.
            if self.station is not None and self.station.isconnected():
                ssid = self.get_ssid()
                if ssid[0] is not None:
                    ip_address = self.get_ip_address()
                    if ip_address is not None and ip_address != '0.0.0.0':
                        return True

        except Exception as ex:
            if ex.__class__.__name__ != 'TimeoutError':
                log.exc(ex, 'Invoking "isconnected" failed')

        return False

    def power_off(self):
        """
        Power off all radio peripherals.
        
        - https://forum.pycom.io/topic/563/disabling-wifi-on-lopy
        - https://github.com/Hiverize/FiPy/commit/b6b15677


        """

        # WiFi
        if self.station:

            if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:

                try:
                    log.info('Turning off WiFi')
                    self.station.deinit()
                except Exception as ex:
                    log.exc(ex, 'Turning off WiFi failed')

            else:   # TODO: other platforms
                pass

    def connect_stations(self, network_names):
        """

        :param network_names: 

        """

        # Prepare information about known WiFi networks.
        network_map = {station['ssid']: station for station in self.stations}

        for network_name in network_names:
            try:
                # All the configuration details for this network.
                # {
                #    'ssid': 'FooBar',
                #    'password': 'SECRET',
                #    'ifconfig': ('192.168.42.42', '255.255.255.0', '192.168.42.1', '192.168.42.1'),
                # }
                network_selected = network_map[network_name]
                if self.connect_station(network_selected):
                    break

            except Exception as ex:
                log.exc(ex, 'WiFi STA: Connecting to "{}" failed'.format(network_name))

        if not self.is_connected():

            self.forget_network(network_name)

            message = 'WiFi STA: Connecting to any network candidate failed'
            description = 'Please check your WiFi configuration for one of the ' \
                          '{} station candidates.'.format(len(network_names))
            log.error('{}. {}'.format(message, description))
            log.warning('Todo: We might want to buffer telemetry data to '
                        'flash memory to be scheduled for transmission later.')

            raise WiFiException(message)

    def connect_station(self, network):
        """

        :param network: 

        """

        network_name = network['ssid']

        log.info('WiFi STA: Preparing connection to network "{}"'.format(network_name))

        auth_mode = None
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            log.info('WiFi STA: Getting auth mode')
            auth_mode = self.get_auth_mode(network_name)
            log.info('WiFi STA: Auth mode is "{}"'.format(auth_mode))

        password = network['password']

        # TODO: Optionally, configure hostname.
        # https://docs.micropython.org/en/latest/library/network.WLAN.html
        # https://github.com/pycom/pycom-micropython-sigfox/pull/165
        # https://forum.pycom.io/topic/3326/new-firmware-release-v1-18-0
        if 'dhcp_hostname' in network:
            if hasattr(self.station, 'config'):
                log.ingo('WiFi STA: Using dhcp_hostname "{}"'.format(network['dhcp_hostname']))
                self.station.config(dhcp_hostname=network['dhcp_hostname'])
            else:
                log.error('Could not set hostname on older MicroPython')

        # Optionally, configure static IP address.
        if 'ifconfig' in network:
            log.info('WiFi STA: Using static network configuration "{}"'.format(network_name))
            self.station.ifconfig(config=network['ifconfig'])

        # Obtain timeout value.
        # TODO: Make default timeout configurable.
        network_timeout = network.get('timeout', 15.0)

        # Connect to WiFi station.
        log.info('WiFi STA: Starting connection to "{}" '
                 'with timeout of {} seconds'.format(network_name, network_timeout))
        self._connect(network_name, password, auth_mode=auth_mode, timeout=network_timeout)

        # After reset, WiFi regularly does not connect.
        # So, let's just try again working around this weird obstacle.
        self.wait_for_connection(2)
        if not self.is_connected():
            self._connect(network_name, password, auth_mode=auth_mode, timeout=network_timeout)

        # Wait for network to arrive.
        self.wait_for_connection(network_timeout)

        # Finally, check for WiFi connection.
        if not self.is_connected():
            raise WiFiException('WiFi STA: Unable to connect to "{}"'.format(network_name))

        # Inform about networking status.
        self.print_short_status()
        self.print_address_status()

        return True

    def _connect(self, ssid, password, auth_mode=None, timeout=None):
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            self.station.connect(ssid, (auth_mode, password), timeout=int(timeout * 1000))
        else:
            self.station.connect(ssid, password)

    def wait_for_connection(self, timeout=15.0):
        """
        Wait for network to arrive.

        :param timeout:  (Default value = 15.0)

        """

        # Set interval how often to poll for WiFi connectivity.
        network_poll_interval = 250

        # How many checks to make.
        checks = int(timeout / (network_poll_interval / 1000.0))

        self.stopwatch.reset()

        do_report = True
        while not self.is_connected():

            delta = self.stopwatch.elapsed()
            eta = timeout - delta

            if checks <= 0 or eta <= 0:
                break

            # Report about the progress each 3 seconds.
            if int(delta) % 3 == 0:
                if do_report:
                    log.info('WiFi STA: Waiting for network to come up within {} seconds'.format(eta))
                    do_report = False
            else:
                do_report = True

            # Save power while waiting.
            machine.idle()

            # Don't busy-wait.
            time.sleep_ms(network_poll_interval)

            checks -= 1

    def scan_stations(self):
        """ """

        # Inquire visible networks.
        log.info("WiFi STA: Scanning for networks")

        # Work around "OSError: Scan operation Failed!" by waiting before `scan`.
        # > I put a time.sleep() of 3sec before wlan.scan(). And the error don't happen.
        # https://forum.pycom.io/topic/5258/firmware-release-v1-20-1/49
        time.sleep(1)

        try:
            stations_available = self.station.scan()
        except Exception as ex:
            log.exc(ex, 'WiFi STA: Scanning for networks failed')
            #if 'Scan operation Failed' in str(ex):
            #    self.station.init()
            return []

        # Collect SSIDs of available stations.
        # (ssid, bssid, channel, RSSI, authmode, hidden)
        try:
            networks_found = frozenset([station.ssid for station in stations_available])
        except AttributeError:
            networks_found = frozenset([station[0] for station in stations_available])

        # Print names/SSIDs of networks found.
        log.info("WiFi STA: Networks available: %s", list(networks_found))

        return stations_available

    def get_ssid(self):
        """ """
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            return self.station.ssid()
        else:
            return self.station.config('essid')

    def get_ip_address(self):
        """ """
        try:
            return self.station.ifconfig()[0]
        except Exception as ex:
            log.exc(ex, 'Unable to get device ip address')

    def get_auth_mode(self, network_name):
        """

        :param network_name: 

        """

        # NVRAM key for storing auth mode per network. Maximum of 15 characters.
        auth_mode_nvs_key = self.auth_mode_nvs_key(network_name.encode())

        # Get WiFi STA auth mode from NVRAM.
        try:
            import pycom
            auth_mode = pycom.nvs_get(auth_mode_nvs_key)
            log.info('WiFi STA: Auth mode from NVRAM with key=%s, value=%s', auth_mode_nvs_key, auth_mode)
        except:
            auth_mode = None

        # Fall back to find out WiFi STA auth mode by network scan.
        if auth_mode is None:

            log.info('WiFi STA: Unknown auth mode for network "%s", invoking WiFi scan', network_name)
            wifi_neighbourhood = self.scan_stations()

            # log.info('WiFi STA: Neighbourhood is %s', wifi_neighbourhood)

            for station in wifi_neighbourhood:

                # (ssid, bssid, channel, RSSI, authmode, hidden)
                log.debug('Station: %s', station)

                try:
                    ssid = station.ssid
                except AttributeError:
                    ssid = station[0].decode()

                if ssid == network_name:
                    try:
                        auth_mode = station.sec
                    except AttributeError:
                        auth_mode = station[4]
                    break

            if auth_mode is None:
                message = 'WiFi STA: Unable to inquire auth mode for network "{}"'.format(network_name)
                log.warning(message)
                raise WiFiException(message)

            try:
                import pycom

                log.info('WiFi STA: Storing auth mode into NVRAM with key=%s, value=%s', auth_mode_nvs_key, auth_mode)
                try:
                    pycom.nvs_set(auth_mode_nvs_key, auth_mode)
                except Exception as ex:
                    log.exc(ex, 'WiFi STA: Storing auth mode into NVRAM failed')

            except ImportError:
                pass

        return auth_mode

    def auth_mode_nvs_key(self, ssid):
        """Hack to get a short representation of a WiFi SSID in order to
        squeeze it into a NVRAM key with a maximum length of 15 characters.
        
        Fixme: Review this.

        :param ssid: 

        """
        import hashlib
        import ubinascii
        try:
            hashfun = hashlib.sha512
        except AttributeError:
            hashfun = hashlib.sha256
        digest = ubinascii.hexlify(hashfun(ssid).digest()).decode()
        identifier = 'wa.{}'.format(digest[15:27])
        return identifier

    def forget_network(self, network_name):
        """

        :param network_name: 

        """
        log.info('WiFi STA: Forgetting NVRAM data for network "{}"'.format(network_name))
        auth_mode_nvs_key = self.auth_mode_nvs_key(network_name)
        try:
            import pycom
            pycom.nvs_erase(auth_mode_nvs_key)
        except:
            pass

    def print_short_status(self):
        """ """
        log.info('WiFi STA: Connected to "{}" with IP address "{}"'.format(self.get_ssid(), self.get_ip_address()))

    def print_address_status(self):
        """ """

        # Get MAC address.
        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Pycom:
            mac_address = self.station.mac()
        else:
            mac_address = self.station.config('mac')

        # Make MAC address human readable.
        mac_address = self.humanize_mac_addresses(mac_address)

        # Get IP address.
        ifconfig = self.station.ifconfig()

        # Display MAC- and IP-address configuration.
        log.info('WiFi STA: Networking address (MAC): %s', mac_address)
        log.info('WiFi STA: Networking address (IP):  %s', ifconfig)

    def humanize_mac_addresses(self, mac):
        """

        :param mac: 

        """
        info = {}
        if hasattr(mac, 'sta_mac'):
            info['sta_mac'] = format_mac_address(binascii.hexlify(mac.sta_mac).decode())
        if hasattr(mac, 'ap_mac'):
            info['ap_mac'] = format_mac_address(binascii.hexlify(mac.ap_mac).decode())
        return info

    def print_metrics(self):
        """ """
        metrics = SystemWiFiMetrics(self.station).read()
        log.info('WiFi STA: Metrics: %s', metrics)


class WiFiException(Exception):
    pass


class SystemWiFiMetrics(AbstractSystemSensor):
    """
    Runtime information from the WiFi bearer.
    """

    def __init__(self, settings, station):
        super().__init__(settings)
        self.station = station

    def read(self):
        """ """

        if self.platform_info.vendor == self.platform_info.MICROPYTHON.Vanilla:
            stats = {}
            #stats['system.wifi.channel'] = self.station.config('channel')

            try:
                stats['system.wifi.rssi'] = self.station.status('rssi')
            except:
                pass

            return stats

        stats = {
            'system.wifi.bandwidth': self.station.bandwidth(),
            'system.wifi.channel': self.station.channel(),
            #'system.wifi.protocol': self.station.wifi_protocol(),
            'system.wifi.max_tx_power': self.station.max_tx_power(),
            #'system.wifi.joined_ap_info': self.station.joined_ap_info(),
        }

        try:
            stats['system.wifi.country'] = self.station.country().country
        except:
            pass

        try:
            stats['system.wifi.rssi'] = self.station.joined_ap_info().rssi
        except:
            pass

        return stats



# -*- coding: utf-8 -*-
#
# Ratrack MicroPython Datalogger
# https://github.com/hiveeyes/terkin-datalogger
#
# (c) 2019 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2019 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
#
from terkin import logging
from terkin.datalogger import TerkinDatalogger
from terkin.telemetry import to_cayenne_lpp
from hiveeyes.sensor_bme280 import BME280Sensor
from hiveeyes.sensor_pytrack import PytrackSensor
from ratrack import convert
from ratrack.sensor_adc import MoistureSensor, WaterlevelSensor
import time

log = logging.getLogger(__name__)


class RatrackDatalogger(TerkinDatalogger):
    """
    Yet another data logger. This time for MicroPython.
    """

    # Naming things.
    name = 'Ratrack MicroPython Datalogger'

    def __init__(self, settings):
        self.data = None
        super().__init__(settings)

    def register_sensors(self):
        """
        Add your sensors here.
        """

        # First, spin up the built-in sensors.
        super().register_sensors()

        # Add some project-specific sensors.
        log.info('Registering Ratrack sensors')

        # Setup the BME280.
        try:
            self.add_bme280_1_sensor()
        except Exception as ex:
            print('INFO:  Skipping bme280 sensor. {}'.format(ex))

        # Setup the BME280.
        try:
            self.add_bme280_sensor()
        except Exception as ex:
            log.exception('Skipping bme280 sensor')

        # Setup the Moisture Sensor.
        try:
            self.add_moisture_sensor()
        except Exception as ex:
            print('INFO:  Skipping Moisture sensor. {}'.format(ex))

        # Setup the Waterlevel Sensor.
        try:
            self.add_waterlevel_sensor()
        except Exception as ex:
            print('INFO:  Skipping Waterlevel sensor. {}'.format(ex))

        # Setup the Pytrack.
        try:
            self.add_pytrack_sensor()
        except Exception as ex:
            log.exception('Skipping Pytrack sensor')

    def add_bme280_sensor(self):
        """
        Setup and register the DS18X20  sensor component with your data logger.
        """

        settings = self.settings.get('sensors.registry.bme280')
        bus = self.sensor_manager.get_bus_by_name(settings['bus'])

        sensor = BME280Sensor()
        sensor.acquire_bus(bus)

        # Start sensor.
        sensor.start()

        # Register with framework.
        self.sensor_manager.register_sensor(sensor)

    def add_bme280_1_sensor(self):
        """
        Setup and register the DS18X20  sensor component with your data logger.
        """

        settings = self.settings.get('sensors.registry.bme280_1')
        bus = self.sensor_manager.get_bus_by_name(settings['bus'])

        sensor = BME280Sensor()
        sensor.acquire_bus(bus)

        # Start sensor.
        sensor.start()

        # Register with framework.
        self.sensor_manager.register_sensor(sensor)

    def add_pytrack_sensor(self):
        """
        Setup and register the Pytrack sensor component with your data logger.
        """

        settings = self.settings.get('sensors.registry.pytrack')
        bus = self.sensor_manager.get_bus_by_name(settings['bus'])

        sensor = PytrackSensor()
        sensor.acquire_bus(bus)

        # Start sensor.
        sensor.start()

        # Register with framework.
        self.sensor_manager.register_sensor(sensor)

    def add_moisture_sensor(self):
        """
        Setup and register the Moisture sensor component with your data logger.
        """

        # Initialize Moisture sensor component.
        settings = self.settings.get('sensors.registry.moisture')

        sensor = MoistureSensor()
        sensor.register_pin('adc_in', settings['pin_adc_in'])
        sensor.register_parameter('scaling', settings['scaling'])

        # Start sensor.
        sensor.start()

        # Register with framework.
        self.sensor_manager.register_sensor(sensor)

    def add_waterlevel_sensor(self):
        """
        Setup and register the Waterlevel sensor component with your data logger.
        """

        # Initialize waterlevel sensor component.
        settings = self.settings.get('sensors.registry.waterlevel')

        sensor = WaterlevelSensor()
        sensor.register_pin('adc_in', settings['pin_adc_in'])

        # Start sensor.
        sensor.start()

        # Register with framework.
        self.sensor_manager.register_sensor(sensor)

    def read_sensors(self):
        self.data = super().read_sensors()
        return self.data

    def loop(self):
        """
        Loop function. Do what I mean.
        """

        # Finally, schedule other system tasks.
        super().loop()
        print('SelfData:', self.data)

        # It's your turn.
        if self.settings.get('networking.lora.antenna_attached'):
            #if self.serialize_and_send():
            if self.to_cayenne_lpp_and_send():
                log.info('[LoRa] Success')


    def to_cayenne_lpp_and_send(self):
        if self.data is None:
            return

            # Create Byte Payload for LoRa.
        try:
            payload = self.to_cayenne_lpp_ratrack()
        except:
            log.exception('[LoRa] Serialization failed')
            return False

            # Send to the wire.
        try:
            lora_received = self.device.networking.lora_send(payload)
            print(lora_received)
        except:
            log.exception('[LoRa] Transmission failed')
            return False

        return True

    def serialize_and_send(self):
        if self.data is None:
            return

        if 'longitude' not in self.data:
            self.data['longitude'] = 0.0
            self.data['latitude'] = 0.0
            self.data['speed'] = 0.0
            self.data['cog'] = 0.0

        # Create Byte Payload for LoRa.
        try:
            payload = convert.create_payload(self.data)
        except:
            log.exception('[LoRa] Serialization failed')
            return False

        # Send to the wire.
        try:
            lora_received = self.device.networking.lora_send(payload)
            print(lora_received)
        except:
            log.exception('[LoRa] Transmission failed')
            return False

        return True

    def to_cayenne_lpp_ratrack(self):
        """
        Serialize plain data dictionary to binary CayenneLPP format.
        """
        data = self.data

        from cayennelpp import LppFrame
        # create empty frame
        frame = LppFrame()


        channel = 1
        # add some sensor data
        if 'temperature' in data:
            try:
                frame.add_temperature(channel, data.get('temperature'))
            except:
                log.exception('[Cayenne] Serialization failed')

        #if 'humidity' in data:
        #    try:
        #        frame.add_humitidy(channel, data.get('add_humidity'))
        #    except:
        #        log.exception('[Cayenne] Serialization failed')

        if 'pressure' in data:
            try:
                frame.add_barometer(channel, data.get('pressure'))
            except:
                log.exception('[Cayenne] Serialization failed')

        if 'speed' in data and 'roll' in data and 'pitch' in data:
            try:
                frame.add_gyrometer(channel, data.get('speed'), data.get('roll'), data.get('pitch'))
            except:
                log.exception('[Cayenne] Serialization failed')

        if 'battery_voltage' in data:
            try:
                frame.add_analog_input(channel, data.get('battery_voltage'))
            except:
                log.exception('[Cayenne] Serialization failed')

        if 'latitude' in data and 'longitude' in data and 'altitude' in data:
            try:
                log.info('GPS to Cayenne')
                frame.add_gps(channel, data.get('latitude'), data.get('longitude'), data.get('altitude'))
            except:
                log.exception('[Cayenne] Serialization failed')

        #channel = 2
        #if 'cog' in data:
        #    try:
        #        frame.add_analog_input(channel, data.get('cog'))
        #    except:
        #        log.exception('[Cayenne] Serialization failed')

        channel = 3
        if 'memfree' in data:
            try:
                frame.add_analog_input(channel, data.get('memfree'))
            except:
                log.exception('[Cayenne] Serialization failed')

        channel = 4
        if 'waterlevel_volt' in data:
            try:
                frame.add_analog_input(channel, data.get('waterlevel_volt'))
            except:
                log.exception('[Cayenne] Serialization failed')

        channel = 5
        if 'moisture_volt' in data:
            try:
                frame.add_analog_input(channel, data.get('moisture_volt'))
            except:
                log.exception('[Cayenne] Serialization failed')

        return frame.bytes()



# -*- coding: utf-8 -*-
# (c) 2019-2020 Jan Hoffmann <jan.hoffmann@bergamsee.de>
# (c) 2019-2020 Andreas Motl <andreas@hiveeyes.org>
# License: GNU General Public License, Version 3
from terkin import logging
from terkin.model import SensorReading
from terkin.sensor import SensorManager, AbstractSensor
from terkin.util import get_platform_info

log = logging.getLogger(__name__)
platform_info = get_platform_info()


def includeme(sensor_manager: SensorManager, sensor_info):
    """
    Create INA219 sensor object.

    :param sensor_manager:
    :param sensor_info:

    :return: sensor_object
    """
    sensor_object = INA219Sensor(settings=sensor_info)
    sensor_bus = sensor_manager.get_bus_by_name(sensor_info.get('bus'))
    sensor_object.acquire_bus(sensor_bus)

    return sensor_object


class INA219Sensor(AbstractSensor):
    """
    A generic INA219 sensor component.
    Used by terkin/datalogger to register and read() from this sensor.
    start() & read() are mandatory.
    """

    def __init__(self, settings=None):

        super().__init__(settings=settings)

        # Can be overwritten by ``.set_address()``.
        self.address = self.settings.get('address', 0x40)

    def start(self):
        """
        Setup the INA219 sensor driver.

        :return:
        """

        # Ensure a bus object exists and is ready.
        self.ensure_bus()

        # Initialize the hardware driver.
        try:

            # MicroPython
            if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:
                from ina219 import INA219
                # TODO: Make "shunt_ohms" configurable. The Adafruit breakout board uses a resistor value of 0.1 Ohms.
                self.driver = INA219(shunt_ohms=0.1, i2c=self.bus.adapter, address=self.address)
                # TODO: Optionally invoke "self.driver.configure()".
                self.driver.configure()

            # Adafruit CircuitPython
            elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:

                # TODO: https://learn.adafruit.com/adafruit-ina219-current-sensor-breakout/python-circuitpython
                # TODO: optional : change configuration to use 32 samples averaging for both bus voltage and shunt voltage

                import adafruit_ina219
                #self.adc_resolution = adafruit_ina219.ADCResolution
                #self.bus_voltage_range = adafruit_ina219.BusVoltageRange
                self.driver = adafruit_ina219.INA219(i2c_bus=self.bus.adapter, addr=self.address)

            else:
                raise NotImplementedError('INA219 driver not implemented on this platform')

            return True

        except Exception as ex:
            log.exc(ex, 'INA219 hardware driver failed')
            return False

    def read(self):
        """
        Read the INA219 sensor.

        :return: SensorReading
        """

        if self.bus is None or self.driver is None:
            return self.SENSOR_NOT_INITIALIZED

        #log.info('Acquire reading from INA219')

        data = {}

        # MicroPython
        if platform_info.vendor in [platform_info.MICROPYTHON.Vanilla, platform_info.MICROPYTHON.Pycom]:

            # Prepare readings.
            values = {
                "bus_voltage": self.driver.voltage(),
                "shunt_voltage": self.driver.shunt_voltage() / 1000.0,
                "current": self.driver.current() / 1000.0,
                "power": self.driver.power() / 1000.0,
            }

            values['psu_voltage'] = values['bus_voltage'] + values['shunt_voltage']

        # Adafruit CircuitPython
        elif platform_info.vendor == platform_info.MICROPYTHON.RaspberryPi:

            values = {

                # Voltage on V- (load side).
                "bus_voltage": self.driver.bus_voltage,

                # Voltage between V+ and V- across the shunt.
                "shunt_voltage": self.driver.shunt_voltage,

                # Current will be acquired in mA, we want A.
                "current": self.driver.current / 1000.0,

                # The power through the load in Watt.
                "power": self.driver.power,

                # Some more values about how the INA219 has been configured.
                # "bus_voltage_range": self.driver.bus_voltage_range,
                # "gain": self.driver.gain,
                # "bus_adc_resolution": self.driver.bus_adc_resolution,
                # "shunt_adc_resolution": self.driver.shunt_adc_resolution,
                # "mode": self.driver.mode,

            }

            values['psu_voltage'] = values['bus_voltage'] + values['shunt_voltage']

            #print(values)
            #print("PSU Voltage:   {:6.3f} V".format(values['bus_voltage'] + values['shunt_voltage']))
            #print("Shunt Voltage: {:9.6f} V".format(values['shunt_voltage']))
            #print("Load Voltage:  {:6.3f} V".format(values['bus_voltage']))
            #print("Current:       {:9.6f} A".format(values['current'] / 1000))

        # Build telemetry payload.
        # TODO: Push this further into the telemetry domain.
        fieldnames = values.keys()
        for name in fieldnames:
            fieldname = self.format_fieldname(name, hex(self.address))
            value = values[name]
            data[fieldname] = value

        if not data:
            log.warning("I2C device {} has no value: {}".format(self.address, data))

        log.debug("I2C data:     {}".format(data))

        reading = SensorReading()
        reading.sensor = self
        reading.data = data

        return reading
